'''Test a single test function on a module with only __init__.py'''
import unittest

from pulsar import send


class TestOneAsync(unittest.TestCase):

    def test_ping_monitor(self):
        future = yield from send('test', 'ping')
        self.assertEqual(future, 'pong')



#!/usr/bin/env python
import sys
import os
from multiprocessing import current_process


def run(**params):
    args = params.get('argv', sys.argv)
    if '--coveralls' in args:
        import pulsar
        from pulsar.utils.path import Path
        from pulsar.apps.test.cov import coveralls

        repo_token = None
        strip_dirs = [Path(pulsar.__file__).parent.parent, os.getcwd()]
        if os.path.isfile('.coveralls-repo-token'):
            with open('.coveralls-repo-token') as f:
                repo_token = f.read().strip()
        coveralls(strip_dirs=strip_dirs, repo_token=repo_token)
        sys.exit(0)
    # Run the test suite
    if '--coverage' in args:
        import coverage
        p = current_process()
        p._coverage = coverage.coverage(data_suffix=True)
        p._coverage.start()
    runtests(**params)


def runtests(test_timeout=None, **params):
    from pulsar.apps.test import TestSuite
    from pulsar.apps.test.plugins import bench, profile
    import pulsar.utils.settings.backend    # noqa

    djangopath = os.path.join(os.path.dirname(__file__),
                              'examples', 'djchat')
    if djangopath not in sys.path:
        sys.path.append(djangopath)
    #
    TestSuite(description='Pulsar Asynchronous test suite',
              modules=('tests',
                       ('examples', 'tests'),
                       ('examples', 'test_*')),
              plugins=(bench.BenchMark(),
                       profile.Profile()),
              pidfile='test.pid',
              test_timeout=test_timeout or 10,
              **params).start()


if __name__ == '__main__':
    run()



'''Tests the test suite loader.'''
import os
import unittest

from pulsar import get_application
from pulsar.apps.test import TestLoader


class TestTestLoader(unittest.TestCase):

    def test_testsuite(self):
        app = yield from get_application('test')
        self.assertTrue(app.cfg.script)
        # self.assertEqual(app.script, sys.argv[0])
        self.assertEqual(os.path.dirname(app.cfg.script), app.root_dir)
        self.assertEqual(app.cfg.modules, ('tests',
                                           ('examples', 'tests'),
                                           ('examples', 'test_*')))

    def test_load_pulsar_tests(self):
        app = yield from get_application('test')
        loader = TestLoader(app.root_dir, app.cfg.modules, app.runner)
        self.assertEqual(loader.modules, [('tests', None, None),
                                          ('examples', 'tests', None),
                                          ('examples', 'test_*', None)])
        modules = dict(loader.testmodules())
        self.assertTrue(modules)
        self.assertFalse('httpbin' in modules)
        self.assertTrue('echo' in modules)
        self.assertTrue('djchat' in modules)
        self.assertTrue('djchat.app' in modules)
        self.assertTrue('djchat.pulse' in modules)
        self.assertTrue('async' in modules)
        self.assertTrue('suite.single' in modules)

    def test_sorted_tags(self):
        app = yield from get_application('test')
        loader = TestLoader(app.root_dir, app.cfg.modules, app.runner)
        modules = list(loader.testmodules())
        self.assertTrue(modules)
        tags = [m[0] for m in modules]
        self.assertEqual(tags, sorted(tags))

    def test_load_tags1(self):
        app = yield from get_application('test')
        loader = TestLoader(app.root_dir, app.cfg.modules, app.runner)
        modules = dict(loader.testmodules(('suite',)))
        self.assertEqual(len(modules), 6)

    def test_load_exclude(self):
        app = yield from get_application('test')
        loader = TestLoader(app.root_dir, app.cfg.modules, app.runner)
        modules = dict(loader.testmodules(
            exclude_tags=('taskqueue', 'apps.pubsub')))
        self.assertTrue(modules)
        for module in modules:
            self.assertTrue('taskqueue' not in module)
            self.assertTrue('apps.pubsub' not in module)



'''MultiFuture coverage'''
import unittest

from pulsar import multi_async, Future


class TestMulti(unittest.TestCase):

    def test_empy_list(self):
        r = multi_async(())
        self.assertTrue(r.done())
        self.assertEqual(r.result(), [])

    def test_empy_dict(self):
        r = multi_async({})
        self.assertTrue(r.done())
        self.assertEqual(r.result(), {})

    def test_multi(self):
        d1 = Future()
        d2 = Future()
        d = multi_async([d1, d2, 'bla'])
        self.assertFalse(d.done())
        d2.set_result('first')
        self.assertFalse(d.done())
        d1.set_result('second')
        result = yield from d
        self.assertEqual(result, ['second', 'first', 'bla'])



import unittest
import asyncio

from pulsar import Protocol


class TestPulsarProtocols(unittest.TestCase):

    def test_pulsar_protocol(self):
        proto = Protocol(asyncio.get_event_loop())
        self.assertEqual(proto.session, 1)
        self.assertFalse(proto.transport)
        self.assertTrue(proto.closed)
        transport = asyncio.Transport()
        proto.connection_made(transport)
        self.assertEqual(proto.transport, transport)



import unittest
import asyncio


def coro(msg):
    future = asyncio.Future()
    future._loop.call_later(1, future.set_result, msg)
    result = yield from future
    return result


class TestCoroutines(unittest.TestCase):

    def test_yield_from(self):
        result = yield from coro('Hi')
        self.assertEqual(result, 'Hi')



import os
from base64 import b64decode
from functools import wraps
import socket
import unittest
import asyncio

import examples

from pulsar import send, SERVER_SOFTWARE, get_event_loop
from pulsar.utils.path import Path
from pulsar.utils.httpurl import iri_to_uri
from pulsar.utils.system import platform
from pulsar.apps.http import (HttpClient, TooManyRedirects, HttpResponse,
                              HTTPError)

__test__ = False


linux = platform.name == 'posix' and not platform.isMacOSX


def dodgyhook(response, exc=None):
    raise ValueError('Dodgy header hook')


def no_tls(f):
    # Don't do the test when tunneling, it causes timeout at times

    @wraps(f)
    def _(self):
        if not self.with_tls:
            return f(self)

    return _


class TestHttpClientBase:
    app = None
    with_httpbin = True
    with_proxy = False
    with_tls = False
    proxy_app = None
    # concurrency is set by the config object unless you set it here
    concurrency = None
    timeout = 10

    @classmethod
    def setUpClass(cls):
        # Create the HttpBin server by sending this request to the arbiter
        from examples.proxyserver.manage import server as pserver
        from examples.httpbin import manage
        concurrency = cls.concurrency or cls.cfg.concurrency
        if cls.with_httpbin:
            server = manage.server
            if cls.with_tls:
                base_path = os.path.abspath(os.path.dirname(manage.__file__))
                key_file = os.path.join(base_path, 'server.key')
                cert_file = os.path.join(base_path, 'server.crt')
            else:
                key_file, cert_file = None, None
            s = server(bind='127.0.0.1:0', concurrency=concurrency,
                       name='httpbin-%s' % cls.__name__.lower(),
                       keep_alive=30, key_file=key_file, cert_file=cert_file,
                       workers=1)
            cfg = yield from send('arbiter', 'run', s)
            cls.app = cfg.app()
            bits = ('https' if cls.with_tls else 'http',) + cfg.addresses[0]
            # bits = (bits[0], 'dynquant.com', '8070')
            cls.uri = '%s://%s:%s/' % bits
        if cls.with_proxy:
            s = pserver(bind='127.0.0.1:0', concurrency=concurrency,
                        name='proxyserver-%s' % cls.__name__.lower())
            cfg = yield from send('arbiter', 'run', s)
            cls.proxy_app = cfg.app()
            cls.proxy_uri = 'http://{0}:{1}'.format(*cfg.addresses[0])
            # cls.proxy_uri = 'http://127.0.0.1:8080'
        cls._client = cls.client()

    @classmethod
    def tearDownClass(cls):
        if cls.app is not None:
            yield from send('arbiter', 'kill_actor', cls.app.name)
        if cls.proxy_app is not None:
            yield from send('arbiter', 'kill_actor', cls.proxy_app.name)

    @classmethod
    def proxies(cls):
        if cls.with_proxy:
            return {'http': cls.proxy_uri,
                    'https': cls.proxy_uri,
                    'ws': cls.proxy_uri,
                    'wss': cls.proxy_uri}

    @classmethod
    def client(cls, loop=None, parser=None, pool_size=2, verify=False,
               **kwargs):
        parser = cls.parser()
        kwargs['proxies'] = cls.proxies()
        return HttpClient(loop=loop, parser=parser, pool_size=pool_size,
                          verify=verify, **kwargs)

    @classmethod
    def parser(cls):
        return None

    @property
    def tunneling(self):
        '''When tunneling, the client needs to perform an extra request.'''
        return int(self.with_proxy and self.with_tls)

    def _check_pool(self, http, response, available=1, processed=1,
                    sessions=1, pools=1):
        # Test the connection pool
        self.assertEqual(len(http.connection_pools), pools)
        if pools:
            pool = http.connection_pools[response.request.key]
            self.assertEqual(http.sessions, sessions)
            self.assertEqual(pool.available, available)
            self.assertEqual(http.requests_processed, processed)

    def _after(self, method, response):
        '''Check for a after_%s % method to test the response.'''
        method = getattr(self, 'after_%s' % method, None)
        if method:
            method(response)

    def httpbin(self, *suffix):
        if suffix:
            return self.uri + '/'.join(suffix)
        else:
            return self.uri

    def after_test_home_page(self, response, processed=1):
        request = response.request
        self.assertEqual(request.scheme, 'https' if self.with_tls else 'http')
        # Only one connection pool,
        # even though the proxy and the connection are for different addresses
        http = response.producer
        self.assertEqual(len(http.connection_pools), 1)
        pool = http.connection_pools[response.request.key]
        self.assertEqual(pool.available, 1)
        self.assertEqual(pool.in_use, 0)
        self.assertEqual(http.sessions, 1)
        self.assertEqual(http.requests_processed, processed)
        self.assertEqual(response._connection._processed, processed)

    def _check_server(self, response):
        self.assertEqual(response.headers['server'], SERVER_SOFTWARE)

    def _test_stream_response(self, siz=3000, rep=10):
        http = self._client
        response = yield from http.get(
            self.httpbin('stream/%d/%d' % (siz, rep)))
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.parser.is_chunked())
        body = response.content
        self.assertEqual(len(body), siz*rep)


class TestHttpClient(TestHttpClientBase, unittest.TestCase):

    def test_home_page(self):
        http = self.client()
        response = yield from http.get(self.httpbin())
        self.assertEqual(str(response), '<Response [200]>')
        self.assertTrue('content-length' in response.headers)
        content = response.content
        size = response.headers['content-length']
        self.assertEqual(len(content), int(size))
        self.assertEqual(response.headers['connection'].lower(), 'keep-alive')
        self._check_server(response)
        self.after_test_home_page(response)
        # Try again
        response = yield from http.get(self.httpbin())
        self.assertEqual(str(response), '<Response [200]>')
        self._check_server(response)
        self.after_test_home_page(response, 2)

    def test_200_get(self):
        http = self.client()
        response = yield from http.get(self.httpbin())
        self.assertEqual(str(response), '<Response [200]>')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.get_status(), '200 OK')
        self.assertTrue(response.content)
        self.assertEqual(response.url, self.httpbin())
        self._check_pool(http, response)
        response = yield from http.get(self.httpbin('get'))
        self.assertEqual(response.status_code, 200)
        self._check_pool(http, response, processed=2)

    def test_200_get_data(self):
        http = self.client()
        response = yield from http.get(self.httpbin('get'),
                                       data={'bla': 'foo'})
        result = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers['content-type'],
                         'application/json; charset=utf-8')
        self.assertEqual(result['args'], {'bla': 'foo'})
        self.assertEqual(response.url,
                         self.httpbin(iri_to_uri('get', {'bla': 'foo'})))
        self._check_pool(http, response)

    def test_200_gzip(self):
        http = self._client
        response = yield from http.get(self.httpbin('gzip'))
        self.assertEqual(response.status_code, 200)
        content = response.json()
        self.assertTrue(content['gzipped'])
        if 'content-encoding' in response.headers:
            self.assertTrue(response.headers['content-encoding'], 'gzip')

    def test_json_post(self):
        http = self._client
        data = {'bla': 'foo',
                'unz': 'whatz',
                'numero': [1, 2]}
        ct = 'application/json'
        response = yield from http.post(self.httpbin('post'),
                                        headers=[('content-type', ct)],
                                        data=data)
        self.assertEqual(response.request.headers['content-type'], ct)
        self.assertEqual(response.status_code, 200)
        result = response.json()
        self.assertTrue(result['args'])
        self.assertEqual(result['args']['numero'], [1, 2])

    def test_upload_images(self):
        http = self._client
        path = Path(examples.__file__).parent.parent
        path = path.join('docs', 'source', '_static')
        files = []
        sent = []
        for name in ('pulsar.png', 'favicon.ico'):
            with open(path.join(name), 'rb') as file:
                image = file.read()
            sent.append(image)
            files.append(('images', (name, image)))
        response = yield from http.put(self.httpbin('upload'), files=files)
        self.assertEqual(response.status_code, 200)
        ct = response.request.headers['content-type']
        self.assertTrue(ct.startswith('multipart/form-data; boundary='))
        data = response.json()
        images = data['files']['images']
        self.assertEqual(len(images), 2)
        for image, s in zip(images, sent):
            image = b64decode(image.encode('utf-8'))
            self.assertEqual(image, s)

    def test_upload_files(self):
        http = self._client
        files = {'test': 'simple file'}
        data = (('bla', 'foo'), ('unz', 'whatz'),
                ('numero', '1'), ('numero', '2'))
        response = yield from http.put(self.httpbin('upload'), data=data,
                                       files=files)
        self.assertEqual(response.status_code, 200)
        ct = response.request.headers['content-type']
        self.assertTrue(ct.startswith('multipart/form-data; boundary='))
        data = response.json()
        self.assertEqual(data['files'], {'test': ['simple file']})
        self.assertEqual(data['args']['numero'], ['1', '2'])

    def test_HttpResponse(self):
        r = HttpResponse(loop=get_event_loop())
        self.assertEqual(r.request, None)
        self.assertEqual(str(r), '<Response [None]>')
        self.assertEqual(r.headers, None)

    def test_client(self):
        http = self.client(max_redirects=5, timeout=33)
        self.assertTrue('accept-encoding' in http.headers)
        self.assertEqual(http.timeout, 33)
        self.assertEqual(http.version, 'HTTP/1.1')
        self.assertEqual(http.max_redirects, 5)
        if self.with_proxy:
            self.assertEqual(http.proxies, {'http': self.proxy_uri,
                                            'https': self.proxy_uri,
                                            'ws': self.proxy_uri,
                                            'wss': self.proxy_uri})

    def test_request_object(self):
        http = self._client
        response = yield from http.get(self.httpbin())
        request = response.request
        self.assertTrue(request.headers)
        self.assertTrue(request.has_header('Connection'))
        self.assertTrue(request.has_header('Accept-Encoding'))
        self.assertTrue(request.has_header('User-Agent'))
        self.assertFalse(request.has_header('foo'))
        self.assertEqual(request.headers.kind, 'client')
        self.assertEqual(request.unredirected_headers.kind, 'client')

    def test_http10(self):
        '''By default HTTP/1.0 close the connection if no keep-alive header
        was passed by the client.
        '''
        http = self.client(version='HTTP/1.0')
        http.headers.clear()
        self.assertEqual(http.version, 'HTTP/1.0')
        response = yield from http.get(self.httpbin())
        self.assertEqual(response.headers['connection'], 'close')
        self.assertEqual(str(response), '<Response [200]>')
        self._check_pool(http, response, available=0)

    def test_http11(self):
        '''By default HTTP/1.1 keep alive the connection if no keep-alive
        header was passed by the client.
        '''
        http = self.client()
        http.headers.clear()
        self.assertEqual(http.version, 'HTTP/1.1')
        response = yield from http.get(self.httpbin())
        self.assertEqual(response.headers['connection'], 'keep-alive')
        self._check_pool(http, response)

    def test_http11_close(self):
        http = self.client()
        self.assertEqual(http.version, 'HTTP/1.1')
        response = yield from http.get(
            self.httpbin(), headers=[('connection', 'close')])
        self.assertEqual(response.headers['connection'], 'close')
        self._check_pool(http, response, available=0)

    def test_post(self):
        data = (('bla', 'foo'), ('unz', 'whatz'),
                ('numero', '1'), ('numero', '2'))
        http = self._client
        response = yield from http.post(self.httpbin('post'),
                                        encode_multipart=False,
                                        data=data)
        self.assertEqual(response.status_code, 200)
        result = response.json()
        self.assertTrue(result['args'])
        self.assertEqual(result['args']['numero'], ['1', '2'])

    def test_400_and_get(self):
        http = self.client()
        response = yield from http.get(self.httpbin('status', '400'))
        self._check_pool(http, response, available=0)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.get_status(), '400 Bad Request')
        self.assertTrue(response.content)
        self.assertRaises(HTTPError, response.raise_for_status)
        # Make sure we only have one connection after a valid request
        response = yield from http.get(self.httpbin('get'))
        self.assertEqual(response.status_code, 200)
        # for tunneling this fails sometimes
        self._check_pool(http, response, sessions=2, processed=2)

    def test_404_get(self):
        http = self._client
        response = yield from http.get(self.httpbin('status', '404'))
        self.assertEqual(response.status_code, 404)
        self.assertTrue(response.headers.has('connection', 'close'))
        self.assertTrue('content-type' in response.headers)
        self.assertTrue(response.content)
        self.assertRaises(HTTPError, response.raise_for_status)

    def test_dodgy_on_header_event(self):
        client = self._client
        response = yield from client.get(self.httpbin(), on_headers=dodgyhook)
        self.assertTrue(response.headers)
        self.assertEqual(response.status_code, 200)

    def test_redirect_1(self):
        http = self.client()
        response = yield from http.get(self.httpbin('redirect', '1'))
        self.assertEqual(response.status_code, 200)
        history = response.history
        self.assertEqual(len(history), 1)
        self.assertTrue(history[0].url.endswith('/redirect/1'))
        self._after('test_redirect_1', response)

    def after_test_redirect_1(self, response):
        redirect = response.history[0]
        self.assertEqual(redirect.connection, response.connection)
        self.assertEqual(response.connection._processed, 2)

    def test_redirect_6(self):
        http = self.client()
        response = yield from http.get(self.httpbin('redirect', '6'))
        self.assertEqual(response.status_code, 200)
        history = response.history
        self.assertEqual(len(history), 6)
        self.assertTrue(history[0].url.endswith('/redirect/6'))
        self._after('test_redirect_6', response)

    def after_test_redirect_6(self, response):
        redirect = response.history[-1]
        self.assertEqual(redirect.connection, response.connection)
        self.assertEqual(response.connection._processed, 7)

    def test_large_response(self):
        http = self._client
        response = yield from http.get(self.httpbin('getsize/600000'))
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(data['size'], 600000)
        self.assertEqual(len(data['data']), 600000)
        self.assertFalse(response.parser.is_chunked())

    def test_too_many_redirects(self):
        http = self._client
        try:
            response = yield from http.get(self.httpbin('redirect', '5'),
                                           max_redirects=2)
        except TooManyRedirects as e:
            response = e.response
        else:
            assert False, 'TooManyRedirects not raised'
        history = response.history
        self.assertEqual(len(history), 2)
        self.assertTrue(history[0].url.endswith('/redirect/5'))
        self.assertTrue(history[1].url.endswith('/redirect/4'))

    def test_post_multipart(self):
        data = (('bla', 'foo'), ('unz', 'whatz'),
                ('numero', '1'), ('numero', '2'))
        http = self._client
        response = yield from http.post(self.httpbin('post'), data=data)
        self.assertEqual(response.status_code, 200)
        result = response.json()
        self.assertTrue(result['args'])
        self.assertEqual(result['args']['numero'], ['1', '2'])

    def test_put(self):
        data = (('bla', 'foo'), ('unz', 'whatz'),
                ('numero', '1'), ('numero', '2'))
        http = self._client
        response = yield from http.put(self.httpbin('put'), data=data)
        self.assertEqual(response.status_code, 200)
        result = response.json()
        self.assertTrue(result['args'])
        self.assertEqual(result['args']['numero'], ['1', '2'])

    def test_patch(self):
        data = (('bla', 'foo'), ('unz', 'whatz'),
                ('numero', '1'), ('numero', '2'))
        http = self._client
        response = yield from http.patch(self.httpbin('patch'),
                                         data=data)
        self.assertEqual(response.status_code, 200)
        result = response.json()
        self.assertTrue(result['args'])
        self.assertEqual(result['args']['numero'], ['1', '2'])

    def test_delete(self):
        data = (('bla', 'foo'), ('unz', 'whatz'),
                ('numero', '1'), ('numero', '2'))
        http = self._client
        response = yield from http.delete(self.httpbin('delete'), data=data)
        self.assertEqual(response.status_code, 200)
        result = response.json()
        self.assertTrue(result['args'])
        self.assertEqual(result['args']['numero'], ['1', '2'])

    def test_response_headers(self):
        http = self._client
        response = yield from http.get(self.httpbin('response_headers'))
        self.assertEqual(response.status_code, 200)
        result = response.json()
        self.assertEqual(result['Transfer-Encoding'], 'chunked')
        parser = response.parser
        self.assertTrue(parser.is_chunked())

    def test_stream_response(self):
        return self._test_stream_response()

    @no_tls
    def test_stream_response_large_chunk(self):
        return self._test_stream_response(100000, 3)

    def test_send_cookie(self):
        http = self._client
        cookies = {'sessionid': 't1', 'cookies_are': 'working'}
        response = yield from http.get(self.httpbin('cookies'),
                                       cookies=cookies)
        self.assertEqual(response.status_code, 200)
        data = response.decode_content()
        self.assertEqual(data['cookies']['sessionid'], 't1')
        self.assertEqual(data['cookies']['cookies_are'], 'working')

    def test_cookie(self):
        http = self._client
        # First set the cookies
        r = yield from http.get(self.httpbin('cookies', 'set', 'bla', 'foo'))
        self.assertEqual(r.status_code, 200)
        self.assertTrue(r.history)
        self.assertTrue(r.history[0].headers['set-cookie'])
        self.assertTrue(http.cookies)
        # Now check if I get them
        r = yield from http.get(self.httpbin('cookies'))
        self.assertEqual(r.status_code, 200)
        self.assertTrue(r.request.unredirected_headers)
        result = r.json()
        self.assertTrue(result['cookies'])
        self.assertEqual(result['cookies']['bla'], 'foo')

    def test_cookie_no_store(self):
        # Try without saving cookies
        http = self.client(store_cookies=False)
        r = yield from http.get(self.httpbin('cookies', 'set', 'bla', 'foo'))
        self.assertEqual(r.status_code, 200)
        self.assertTrue(r.history)
        self.assertTrue(r.history[0].headers['set-cookie'])
        r = yield from http.get(self.httpbin('cookies'))
        self.assertEqual(r.status_code, 200)
        result = r.json()
        self.assertFalse(result['cookies'])

    def test_basic_authentication(self):
        http = self._client
        r = yield from http.get(self.httpbin('basic-auth/bla/foo'))
        # The response MUST include a WWW-Authenticate header field
        self.assertEqual(r.status_code, 401)
        r = yield from http.get(self.httpbin('basic-auth/bla/foo'),
                                auth=('bla', 'foo'))
        self.assertEqual(r.status_code, 200)

    def test_digest_authentication(self):
        http = self.client()
        r = yield from http.get(self.httpbin(
            'digest-auth/luca/bla/auth'))
        self.assertEqual(r.status_code, 401)
        http.add_digest_authentication('luca', 'bla')
        r = yield from http.get(self.httpbin(
            'digest-auth/luca/bla/auth'))
        self.assertEqual(r.status_code, 200)

    def test_missing_host_400(self):
        http = self._client

        def remove_host(response, exc=None):
            r = response.request
            self.assertTrue(r.has_header('host'))
            response.request.remove_header('host')
            self.assertFalse(r.has_header('host'))

        response = yield from http.get(self.httpbin(),
                                       pre_request=remove_host)
        self.assertEqual(response.status_code, 400)

    def test_missing_host_10(self):
        http = self.client(version='HTTP/1.0')

        def remove_host(response, exc=None):
            request = response.request
            if not hasattr(request, '_test_host'):
                request._test_host = request.remove_header('host')

        response = yield from http.get(self.httpbin(),
                                       pre_request=remove_host)
        self.assertEqual(response.status_code, 200)
        request = response.request
        self.assertFalse(request.has_header('host'))
        self.assertTrue(request._test_host)

    def test_expect(self):
        http = self._client
        data = (('bla', 'foo'), ('unz', 'whatz'),
                ('numero', '1'), ('numero', '2'))
        response = yield from http.post(self.httpbin('post'), data=data,
                                        wait_continue=True)
        self.assertEqual(response.status_code, 200)
        result = response.json()
        self.assertTrue(result['args'])
        self.assertEqual(result['args']['numero'], ['1', '2'])

    def test_expect_fail(self):
        '''This is an important test for the proxy server example.
        The expect-continue must be handled by the upstream server which in
        this case refuses the continue.'''
        http = self._client
        data = (('bla', 'foo'), ('unz', 'whatz'),
                ('numero', '1'), ('numero', '2'))
        response = yield from http.post(self.httpbin('expect'), data=data,
                                        wait_continue=True)
        self.assertEqual(response.status_code, 417)

    def test_media_root(self):
        http = self._client
        response = yield from http.get(self.httpbin('media/'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers['content-type'],
                         'text/html; charset=utf-8')

    def test_media_file(self):
        http = self._client
        response = yield from http.get(self.httpbin('media/httpbin.js'))
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.headers['content-type'] in
                        ('application/javascript',
                         'application/x-javascript'))
        self.assertTrue(int(response.headers['content-length']) > 0)
        modified = response.headers.get('Last-modified')
        self.assertTrue(modified)
        #
        # Test if modified since
        response = yield from http.get(
            self.httpbin('media/httpbin.js'),
            headers=[('If-modified-since', modified)])
        self.assertEqual(response.status_code, 304)
        self.assertFalse('Content-length' in response.headers)

    def test_http_get_timeit(self):
        N = 10
        client = self._client
        bench = yield from client.timeit('get', N, self.httpbin('get'),
                                         data={'bla': 'foo'})
        self.assertTrue(bench.taken)
        self.assertEqual(len(bench.result), N)
        for r in bench.result:
            self.assertEqual(r.status_code, 200)

    def test_send_files(self):
        client = self._client
        files = {'test': 'simple file'}
        data = (('bla', 'foo'), ('unz', 'whatz'),
                ('numero', '1'), ('numero', '2'))
        response = yield from client.post(self.httpbin('post'), data=data,
                                          files=files)
        self.assertEqual(response.status_code, 200)
        ct = response.request.headers['content-type']
        self.assertTrue(ct.startswith('multipart/form-data; boundary='))
        data = response.json()
        self.assertEqual(data['files'], {'test': ['simple file']})
        self.assertEqual(data['args']['numero'], ['1', '2'])

    def test_send_images(self):
        path = Path(examples.__file__).parent.parent
        path = path.join('docs', 'source', '_static')
        files = []
        sent = []
        for name in ('pulsar.png', 'favicon.ico'):
            with open(path.join(name), 'rb') as file:
                image = file.read()
            sent.append(image)
            files.append(('images', (name, image)))
        client = self._client
        response = yield from client.post(self.httpbin('post'), files=files)
        self.assertEqual(response.status_code, 200)
        ct = response.request.headers['content-type']
        self.assertTrue(ct.startswith('multipart/form-data; boundary='))
        data = response.json()
        images = data['files']['images']
        self.assertEqual(len(images), 2)
        for image, s in zip(images, sent):
            image = b64decode(image.encode('utf-8'))
            self.assertEqual(image, s)

    def test_bench_json(self):
        http = self._client
        response = yield from http.get(self.httpbin('json'))
        self.assertEqual(response.headers['content-type'],
                         'application/json; charset=utf-8')
        result = response.decode_content()
        self.assertEqual(result, {'message': 'Hello, World!'})

    def test_bench_text(self):
        http = self._client
        response = yield from http.get(self.httpbin('plaintext'))
        self.assertEqual(response.headers['content-type'],
                         'text/plain; charset=utf-8')
        result = response.decode_content()
        self.assertEqual(result, 'Hello, World!')

    def test_pool_200(self):
        N = 6
        http = self.client(pool_size=2)
        bench = yield from http.timeit('get', N, self.httpbin())
        self.assertEqual(len(bench.result), N)
        for response in bench.result:
            self.assertEqual(str(response), '<Response [200]>')
            self.assertTrue('content-length' in response.headers)
        self.assertEqual(len(http.connection_pools), 1)
        pool = tuple(http.connection_pools.values())[0]
        self.assertEqual(pool.pool_size, 2)
        self.assertEqual(pool.in_use, 0)
        self.assertEqual(pool.available, 2)

    def test_pool_400(self):
        N = 6
        http = self.client(pool_size=2)
        bench = yield from http.timeit('get', N, self.httpbin('status', '400'))
        self.assertEqual(len(bench.result), N)
        for response in bench.result:
            self.assertEqual(str(response), '<Response [400]>')
            self.assertTrue('content-length' in response.headers)
        self.assertEqual(len(http.connection_pools), 1)
        pool = tuple(http.connection_pools.values())[0]
        self.assertEqual(pool.pool_size, 2)
        self.assertEqual(pool.in_use, 0)
        self.assertEqual(pool.available, 0)

    def test_415(self):
        http = self._client
        response = yield from http.get(
            self.httpbin(''), headers=[('accept', 'application/json')])
        self.assertEqual(response.status_code, 415)

    @unittest.skipUnless(linux, 'Test in lunux platform only')
    def test_servername(self):
        http = self.client()
        http.headers.remove_header('host')
        self.assertNotIn('host', http.headers)
        http.headers['host'] = 'fakehost'
        response = yield from http.get(self.httpbin('servername'))
        self.assertEqual(response.status_code, 200)
        result = response.decode_content().strip()
        # result should resolve to loopback address
        ips = {sockaddr[0] for _, _, _, _, sockaddr in
               socket.getaddrinfo(result, None)}
        self.assertTrue({'127.0.0.1', '::1'} & ips)

    def test_raw_property(self):
        http = self._client
        response = yield from http.get(self.httpbin('plaintext'))
        raw = response.raw
        self.assertEqual(raw._response, response)
        yield from self.async.assertEqual(raw.read(), b'')

    def test_stream_dont_stream(self):
        http = self._client
        response = yield from http.get(self.httpbin('plaintext'), stream=True)
        yield from response.on_finished
        self.async.assertEqual(response.text(), 'Hello, World!')

    def test_raw_stream(self):
        http = self._client
        response = yield from http.get(self.httpbin('plaintext'), stream=True)
        raw = response.raw
        self.assertEqual(raw._response, response)
        yield from self.async.assertEqual(raw.read(), b'Hello, World!')
        self.assertTrue(raw.done)
        yield from self.async.assertEqual(raw.read(), b'')

    @no_tls
    def test_raw_stream_large(self):
        http = self._client
        url = self.httpbin('stream/100000/3')
        response = yield from http.get(url, stream=True)
        raw = response.raw
        self.assertEqual(raw._response, response)
        data = yield from raw.read()
        self.assertTrue(len(data), 300000)

    def test_post_iterator(self):
        http = self._client
        fut = asyncio.Future()

        def gen():
            yield b'a'*100
            yield fut
            yield b'z'*100

        result = b'f'*100
        fut._loop.call_later(0.5, fut.set_result, result)
        response = yield from http.post(self.httpbin('post_chunks'),
                                        data=gen())
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.content), 300)
        self.assertEqual(response.headers['content-type'],
                         response.request.headers['content-type'])



import unittest

import pulsar


class Context:

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._result = value
        return True


class TestApi(unittest.TestCase):

    def test_with_statement(self):
        with Context() as c:
            yield None
            yield None
            raise ValueError
        self.assertIsInstance(c._result, ValueError)

    def test_get_proxy(self):
        self.assertRaises(ValueError, pulsar.get_proxy, 'shcbjsbcjcdcd')
        self.assertEqual(pulsar.get_proxy('shcbjsbcjcdcd', safe=True), None)

    def test_bad_concurrency(self):
        # bla concurrency does not exists
        yield from self.async.assertRaises(ValueError,
                                           pulsar.spawn, kind='bla')

    def test_actor_coverage(self):
        '''test case for coverage'''
        yield from self.async.assertRaises(pulsar.CommandNotFound,
                                           pulsar.send, 'arbiter',
                                           'sjdcbhjscbhjdbjsj', 'bla')



import unittest

from pulsar.apps.http import HttpClient, HttpRequest, parse_qsl


class TestClientCornerCases(unittest.TestCase):

    def test_headers(self):
        headers = HttpClient.DEFAULT_HTTP_HEADERS
        self.assertEqual(len(headers), 3)
        accept = headers['accept-encoding']
        self.assertTrue('gzip' in accept)
        self.assertTrue('deflate' in accept)

    def test_override_headers(self):
        headers = {'Accept': 'application/json, text/plain; q=0.8',
                   'content-type': 'application/json'}
        client = HttpClient(headers=headers)
        self.assertEqual(client.headers['accept'],
                         'application/json, text/plain; q=0.8')
        self.assertEqual(client.headers['content-type'], 'application/json')

    def test_urlparams(self):
        http = HttpClient()
        urlparams = {'page': 2, 'key': 'foo'}
        request = HttpRequest(http, 'http://bla.com?k=6', 'post',
                              urlparams=urlparams)
        params = parse_qsl(request.query)
        self.assertEqual(len(params), 3)



from pulsar import HAS_C_EXTENSIONS
from pulsar.apps.test import check_server
from pulsar.apps.data import RedisScript

from .pulsards import unittest, RedisCommands, create_store
from .lock import RedisLockTests


OK = check_server('redis')


@unittest.skipUnless(OK, 'Requires a running Redis server')
class RedisDbTest(RedisCommands, RedisLockTests):
    pass


@unittest.skipUnless(OK, 'Requires a running redis server')
class TestRedisStore(RedisDbTest, unittest.TestCase):
    store = None

    @classmethod
    def setUpClass(cls):
        addr = cls.cfg.redis_server
        if not addr.startswith('redis://'):
            addr = 'redis://%s' % cls.cfg.redis_server
        namespace = cls.__name__.lower()
        cls.store = create_store(addr, pool_size=3, namespace=namespace)
        cls.client = cls.store.client()

    def test_script(self):
        script = RedisScript("return 1")
        self.assertFalse(script.sha)
        self.assertTrue(script.script)
        result = yield from script(self.client)
        self.assertEqual(result, 1)
        self.assertTrue(script.sha)
        self.assertTrue(script.sha in self.client.store.loaded_scripts)
        result = yield from script(self.client)
        self.assertEqual(result, 1)

    def test_eval(self):
        result = yield from self.client.eval('return "Hello"')
        self.assertEqual(result, b'Hello')
        result = yield from self.client.eval("return {ok='OK'}")
        self.assertEqual(result, b'OK')

    def test_eval_with_keys(self):
        result = yield from self.client.eval("return {KEYS, ARGV}",
                                             ('a', 'b'),
                                             ('first', 'second', 'third'))
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], [b'a', b'b'])
        self.assertEqual(result[1], [b'first', b'second', b'third'])


@unittest.skipUnless(OK and HAS_C_EXTENSIONS, 'Requires cython extensions')
class TestRedisStorePyParser(TestRedisStore):
    pass



import asyncio

from pulsar.apps.data import LockError


class RedisLockTests:

    def test_lock(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        lock = self.client.lock(key)
        self.assertEqual(lock.name, key)
        self.assertFalse(lock.token)

        yield from eq(lock.acquire(), True)
        self.assertTrue(lock.token)
        yield from eq(self.client.get(key), lock.token)
        # assert sr.ttl('foo') == -1
        yield from eq(lock.release(), True)
        yield from eq(self.client.get(key), None)

    def test_competing_locks(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        lock1 = self.client.lock(key)
        lock2 = self.client.lock(key)
        yield from eq(lock1.acquire(), True)
        yield from eq(lock2.acquire(), False)
        yield from eq(lock1.release(), True)
        yield from eq(lock2.acquire(), True)
        yield from eq(lock1.acquire(), False)
        yield from eq(lock2.release(), True)

    def test_timeout(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        lock = self.client.lock(key, timeout=10)
        yield from eq(lock.acquire(), True)
        ttl = yield from self.client.ttl(lock.name)
        self.assertTrue(2 < ttl <= 10)
        yield from eq(lock.release(), True)

    def test_float_timeout(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        lock = self.client.lock(key, timeout=9.5)
        yield from eq(lock.acquire(), True)
        ttl = yield from self.client.pttl(lock.name)
        self.assertTrue(4000 < ttl <= 9500)
        yield from eq(lock.release(), True)

    def test_blocking_timeout(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        lock1 = self.client.lock(key)
        lock2 = self.client.lock(key, blocking=0.5)
        yield from eq(lock1.acquire(), True)
        start = lock1._loop.time()
        yield from eq(lock2.acquire(), False)
        self.assertTrue(lock1._loop.time() - start > 0.5)
        yield from eq(lock1.release(), True)

    def test_blocking_timeout_acquire(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        lock1 = self.client.lock(key)
        lock2 = self.client.lock(key, blocking=5)
        yield from eq(lock1.acquire(), True)
        asyncio.async(self._release(lock1, 0.5))
        start = lock2._loop.time()
        yield from eq(lock2.acquire(), True)
        self.assertTrue(5 > lock2._loop.time() - start > 0.5)
        yield from eq(lock2.release(), True)

    def _release(self, lock, time):
        yield from asyncio.sleep(time)
        yield from self.async.assertEqual(lock.release(), True)

    def test_high_sleep_raises_error(self):
        "If sleep is higher than timeout, it should raise an error"
        self.assertRaises(LockError, self.client.lock, 'foo',
                          blocking=1, sleep=2)

    def test_releasing_lock_no_longer_owned_raises_error(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        lock = self.client.lock(key)
        yield from eq(lock.acquire(), True)
        # manually change the token
        yield from self.client.set(key, 'a')
        yield from self.async.assertRaises(LockError, lock.release)
        # even though we errored, the token is still cleared
        self.assertEqual(lock.token, None)



'''Tests the tools and utilities in pulsar.utils.'''
import unittest

from pulsar.utils.log import lazy_string


@lazy_string
def blabla(n):
    return 'AAAAAAAAAAAAAAAAAAAA %s' % n


class TestTextUtils(unittest.TestCase):

    def testLazy(self):
        r = blabla(3)
        self.assertEqual(r.value, None)
        v = str(r)
        self.assertEqual(v, 'AAAAAAAAAAAAAAAAAAAA 3')
        self.assertEqual(r.value, v)



import socket
import unittest
from unittest import mock

from pulsar.utils.internet import (parse_address, parse_connection_string,
                                   close_socket, format_address)


class TestParseAddress(unittest.TestCase):

    def test_parse_ipv4(self):
        address = parse_address('127.0.0.1')
        self.assertEqual(address, ('127.0.0.1', 8000))
        address = parse_address('127.0.0.1', 8060)
        self.assertEqual(address, ('127.0.0.1', 8060))

    def test_parse_ipv6(self):
        address = parse_address('[::1]')
        self.assertEqual(address, ('::1', 8000))
        address = parse_address('[::1]:8070')
        self.assertEqual(address, ('::1', 8070))

    def test_parse_error(self):
        self.assertRaises(ValueError, parse_address, ())
        self.assertRaises(ValueError, parse_address, (1,))
        self.assertRaises(ValueError, parse_address, (1, 2, 3))
        self.assertRaises(ValueError, parse_address, '127.0.0.1:bla')


class TestParseConnectionString(unittest.TestCase):

    def test_parse_tcp(self):
        scheme, address, params = parse_connection_string('127.0.0.1:8050')
        self.assertEqual(scheme, '')
        self.assertEqual(address, ('127.0.0.1', 8050))
        self.assertEqual(params, {})

    def test_parse_tcp_default(self):
        scheme, address, params = parse_connection_string('127.0.0.1', 8095)
        self.assertEqual(scheme, '')
        self.assertEqual(address, ('127.0.0.1', 8095))
        self.assertEqual(params, {})

    def test_parse_unix(self):
        scheme, address, params = parse_connection_string('unix:bla.foo')
        self.assertEqual(scheme, '')
        self.assertEqual(address, 'bla.foo')
        self.assertEqual(params, {})

    def test_parse_unix_with_scheme(self):
        scheme, address, params = parse_connection_string(
            'redis://unix:bla.foo')
        self.assertEqual(scheme, 'redis')
        self.assertEqual(address, 'bla.foo')
        self.assertEqual(params, {})

    def test_parse_tcp_with_scheme_and_params(self):
        scheme, address, params = parse_connection_string('redis://:6439?db=3')
        self.assertEqual(scheme, 'redis')
        self.assertEqual(address, ('', 6439))
        self.assertEqual(params, {'db': '3'})

    def test_parse_tcp_with_http_and_params(self):
        scheme, address, params = parse_connection_string(
            'http://:6439?db=3&bla=foo')
        self.assertEqual(scheme, 'http')
        self.assertEqual(address, ('', 6439))
        self.assertEqual(params, {'db': '3', 'bla': 'foo'})

    def test_parse_tcp_with_https_and_params(self):
        scheme, address, params = parse_connection_string(
            'https://127.0.0.1:6439?db=3&bla=foo')
        self.assertEqual(scheme, 'https')
        self.assertEqual(address, ('127.0.0.1', 6439))
        self.assertEqual(params, {'db': '3', 'bla': 'foo'})


class TestMisc(unittest.TestCase):

    def test_close_socket(self):
        close_socket(None)
        sock = mock.Mock()
        sock.configure_mock(**{'shutdown.side_effect': TypeError,
                               'close.side_effect': TypeError})
        close_socket(sock)
        sock.shutdown.assert_called_with(socket.SHUT_RDWR)
        sock.close.assert_called_with()

    def test_format_address(self):
        self.assertRaises(ValueError, format_address, (1,))
        self.assertRaises(ValueError, format_address, (1, 2, 3))
        self.assertRaises(ValueError, format_address, (1, 2, 3, 4, 5))
        self.assertEqual(format_address(1), '1')



import binascii
import time
import unittest
import asyncio
import datetime

import pulsar
from pulsar.utils.string import random_string
from pulsar.utils.structures import Zset
from pulsar.apps.ds import PulsarDS, redis_parser, ResponseError
from pulsar.apps.data import create_store


class Listener:

    def __init__(self):
        self._messages = asyncio.Queue()

    def __call__(self, channel, message):
        self._messages.put_nowait((channel, message))

    def get(self):
        return self._messages.get()


class StringProtocol:

    def encode(self, message):
        return message

    def decode(self, message):
        return message.decode('utf-8')


class StoreMixin:
    redis_py_parser = False

    @classmethod
    def create_store(cls, address, namespace=None, pool_size=2, **kw):
        if cls.redis_py_parser:
            kw['parser_class'] = redis_parser(True)
        if not namespace:
            namespace = cls.randomkey(6).lower()
        return create_store(address, namespace=namespace,
                            pool_size=pool_size, **kw)

    @classmethod
    def randomkey(cls, length=None):
        return random_string(min_length=length, max_length=length)

    def _remove_and_push(self, key, rem=1):
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.delete(key), rem)
        yield from eq(c.rpush(key, 'bla'), 1)

    def _remove_and_sadd(self, key, rem=1):
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.delete(key), rem)
        yield from eq(c.sadd(key, 'bla'), 1)


class RedisCommands(StoreMixin):

    def test_store(self):
        store = self.store
        self.assertTrue(store.namespace)

    ###########################################################################
    #    KEYS
    def test_dump_restore(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.dump(key), None)
        yield from eq(c.set(key, 'hello'), True)
        value = yield from c.dump(key)
        self.assertTrue(value)
        yield from self.async.assertRaises(
            ResponseError, c.restore, key, 0, 'bla')
        yield from eq(c.restore(key+'2', 0, value), True)
        yield from eq(c.get(key+'2'), b'hello')

    def test_exists(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.exists(key), False)
        yield from eq(c.set(key, 'hello'), True)
        yield from eq(c.exists(key), True)
        yield from eq(c.delete(key), 1)

    def test_expire_persist_ttl(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from self.async.assertRaises(ResponseError, c.expire, key, 'bla')
        yield from eq(c.expire(key, 1), False)
        yield from eq(c.set(key, 1), True)
        yield from eq(c.expire(key, 10), True)
        ttl = yield from c.ttl(key)
        self.assertTrue(ttl > 0 and ttl <= 10)
        yield from eq(c.persist(key), True)
        yield from eq(c.ttl(key), -1)
        yield from eq(c.persist(key), False)

    def test_expireat(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from self.async.assertRaises(
            ResponseError, c.expireat, key, 'bla')
        t = int(time.time() + 3)
        yield from eq(c.expireat(key, t), False)
        yield from eq(c.set(key, 1), True)
        t = int(time.time() + 10)
        yield from eq(c.expireat(key, t), True)
        ttl = yield from c.ttl(key)
        self.assertTrue(ttl > 0 and ttl <= 10)
        yield from eq(c.persist(key), True)
        yield from eq(c.ttl(key), -1)
        yield from eq(c.persist(key), False)

    def test_keys(self):
        key = self.randomkey()
        keya = '%s_a' % key
        keyb = '%s_a' % key
        keyc = '%sc' % key
        c = self.client
        eq = self.async.assertEqual
        keys_with_underscores = set([keya.encode('utf-8'),
                                     keyb.encode('utf-8')])
        keys = keys_with_underscores.union(set([keyc.encode('utf-8')]))
        yield from eq(c.mset(keya, 1, keyb, 2, keyc, 3), True)
        k1 = yield from c.keys('%s_*' % key)
        k2 = yield from c.keys('%s*' % key)
        self.assertEqual(set(k1), keys_with_underscores)
        self.assertEqual(set(k2), keys)

    def test_move(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        db = 3 if c.store.database == 4 else 4
        yield from self.async.assertRaises(ResponseError, c.move, key, 'bla')
        yield from eq(c.move(key, db), False)
        yield from eq(c.set(key, 'ciao'), True)
        yield from eq(c.move(key, db), True)
        s2 = self.create_store(self.store.dns, database=db)
        c2 = s2.client()
        yield from eq(c2.get(key), b'ciao')
        yield from eq(c.exists(key), False)
        yield from eq(c.set(key, 'foo'), True)
        yield from eq(c.move(key, db), False)
        yield from eq(c.exists(key), True)

    def __test_randomkey(self):
        # TODO: this test fails sometimes
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.mset(key, 1, key+'a', 2, key+'b', 3), True)
        key = yield from c.randomkey()
        yield from eq(c.exists(key), True)

    def test_rename_renamenx(self):
        key = self.randomkey()
        des = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from self.async.assertRaises(ResponseError, c.rename, key, des)
        yield from eq(c.set(key, 'hello'), True)
        yield from self.async.assertRaises(ResponseError, c.rename, key, key)
        yield from eq(c.rename(key, des), True)
        yield from eq(c.exists(key), False)
        yield from eq(c.get(des), b'hello')
        yield from eq(c.set(key, 'ciao'), True)
        yield from eq(c.renamenx(key, des), False)
        yield from eq(c.renamenx(key, des+'a'), True)
        yield from eq(c.exists(key), False)

    ###########################################################################
    #    BAD REQUESTS
    # def test_no_command(self):
    #     yield from self.async.assertRaises(ResponseError, self.store.execute)

    # def test_bad_command(self):
    #     yield from self.async.assertRaises(ResponseError, self.store.execute,
    #                                   'foo')
    ###########################################################################
    #    STRINGS
    def test_append(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.append(key, 'a1'), 2)
        yield from eq(c.get(key), b'a1')
        yield from eq(c.append(key, 'a2'), 4)
        yield from eq(c.get(key), b'a1a2')
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(ResponseError, c.append, key, 'g')

    def test_bitcount(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.bitcount(key), 0)
        yield from eq(c.setbit(key, 5, 1), 0)
        yield from eq(c.bitcount(key), 1)
        yield from eq(c.setbit(key, 6, 1), 0)
        yield from eq(c.bitcount(key), 2)
        yield from eq(c.setbit(key, 5, 0), 1)
        yield from eq(c.bitcount(key), 1)
        yield from eq(c.setbit(key, 9, 1), 0)
        yield from eq(c.setbit(key, 17, 1), 0)
        yield from eq(c.setbit(key, 25, 1), 0)
        yield from eq(c.setbit(key, 33, 1), 0)
        yield from eq(c.bitcount(key), 5)
        yield from eq(c.bitcount(key, 0, -1), 5)
        yield from eq(c.bitcount(key, 2, 3), 2)
        yield from eq(c.bitcount(key, 2, -1), 3)
        yield from eq(c.bitcount(key, -2, -1), 2)
        yield from eq(c.bitcount(key, 1, 1), 1)
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(ResponseError, c.bitcount, key)

    def test_bitop_not_empty_string(self):
        key = self.randomkey()
        des = key + 'd'
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.set(key, ''), True)
        yield from eq(c.bitop('not', des, key), 0)
        yield from eq(c.get(des), None)

    def test_bitop_not(self):
        key = self.randomkey()
        des = key + 'd'
        c = self.client
        eq = self.async.assertEqual
        test_str = b'\xAA\x00\xFF\x55'
        correct = ~0xAA00FF55 & 0xFFFFFFFF
        yield from eq(c.set(key, test_str), True)
        yield from eq(c.bitop('not', des, key), 4)
        result = yield from c.get(des)
        self.assertEqual(int(binascii.hexlify(result), 16), correct)

    def test_bitop_not_in_place(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        test_str = b'\xAA\x00\xFF\x55'
        correct = ~0xAA00FF55 & 0xFFFFFFFF
        yield from eq(c.set(key, test_str), True)
        yield from eq(c.bitop('not', key, key), 4)
        result = yield from c.get(key)
        assert int(binascii.hexlify(result), 16) == correct

    def test_bitop_single_string(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        test_str = b'\x01\x02\xFF'
        yield from eq(c.set(key, test_str), True)
        yield from eq(c.bitop('and', key+'1', key), 3)
        yield from eq(c.bitop('or', key+'2', key), 3)
        yield from eq(c.bitop('xor', key+'3', key), 3)
        yield from eq(c.get(key + '1'), test_str)
        yield from eq(c.get(key + '2'), test_str)
        yield from eq(c.get(key + '3'), test_str)

    def test_bitop_string_operands(self):
        c = self.client
        eq = self.async.assertEqual
        key1 = self.randomkey()
        key2 = key1 + '2'
        des1 = key1 + 'd1'
        des2 = key1 + 'd2'
        des3 = key1 + 'd3'
        yield from eq(c.set(key1, b'\x01\x02\xFF\xFF'), True)
        yield from eq(c.set(key2, b'\x01\x02\xFF'), True)
        yield from eq(c.bitop('and', des1, key1, key2), 4)
        yield from eq(c.bitop('or', des2, key1, key2), 4)
        yield from eq(c.bitop('xor', des3, key1, key2), 4)
        res1 = yield from c.get(des1)
        res2 = yield from c.get(des2)
        res3 = yield from c.get(des3)
        self.assertEqual(int(binascii.hexlify(res1), 16), 0x0102FF00)
        self.assertEqual(int(binascii.hexlify(res2), 16), 0x0102FFFF)
        self.assertEqual(int(binascii.hexlify(res3), 16), 0x000000FF)

    def test_decr(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.decr(key), -1)
        yield from eq(c.get(key), b'-1')
        yield from eq(c.decr(key), -2)
        yield from eq(c.get(key), b'-2')
        yield from eq(c.decr(key, 5), -7)
        yield from eq(c.get(key), b'-7')

    def test_getbit(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.getbit(key, 5), 0)
        yield from eq(c.setbit(key, 5, 1), 0)
        yield from eq(c.getbit(key, 5), 1)
        yield from self.async.assertRaises(ResponseError, c.getbit, key, -1)
        yield from eq(c.getbit(key, 4), 0)
        yield from eq(c.setbit(key, 4, 1), 0)
        # set bit 4
        yield from eq(c.getbit(key, 4), 1)
        yield from eq(c.getbit(key, 5), 1)
        # set bit 5 again
        yield from eq(c.setbit(key, 5, 1), 1)
        yield from eq(c.getbit(key, 5), 1)
        #
        yield from eq(c.getbit(key, 30), 0)
        #
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(ResponseError, c.getbit, key, 1)

    def test_getrange(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.getrange(key, 0, 0), b'')
        yield from eq(c.set(key, 'Hello there'), True)
        yield from eq(c.getrange(key, 0, 0), b'H')
        yield from eq(c.getrange(key, 0, 4), b'Hello')
        yield from eq(c.getrange(key, 5, 5), b' ')
        yield from eq(c.getrange(key, 20, 25), b'')
        yield from eq(c.getrange(key, -5, -1), b'there')
        yield from self.async.assertRaises(
            ResponseError, c.getrange, key, 1, 'b')
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(
            ResponseError, c.getrange, key, 1, 2)

    def test_getset(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.getset(key, 'foo'), None)
        yield from eq(c.getset(key, 'bar'), b'foo')
        yield from eq(c.get(key), b'bar')

    def test_get(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from c.set(key, 'foo')
        yield from eq(c.get(key), b'foo')
        yield from eq(c.get('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'), None)

    def test_incr(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.incr(key), 1)
        yield from eq(c.get(key), b'1')
        yield from eq(c.incr(key), 2)
        yield from eq(c.get(key), b'2')
        yield from eq(c.incr(key, 5), 7)
        yield from eq(c.get(key), b'7')

    def test_incrby(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.incrby(key), 1)
        yield from eq(c.incrby(key, 4), 5)
        yield from eq(c.get(key), b'5')

    def test_incrbyfloat(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.incrbyfloat(key), 1.0)
        yield from eq(c.get(key), b'1')
        yield from eq(c.incrbyfloat(key, 1.1), 2.1)
        yield from eq(c.get(key), b'2.1')

    def test_mget(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        key3 = key2 + 'y'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.set(key1, 'foox'), True)
        yield from eq(c.set(key2, 'fooxx'), True)
        yield from eq(c.mget(key1, key2, key3), [b'foox', b'fooxx', None])

    def test_msetnx(self):
        key1 = self.randomkey()
        key2 = self.randomkey()
        key3 = self.randomkey()
        key4 = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.msetnx(key1, '1', key2, '2', key3, '3'), True)
        yield from eq(c.msetnx(key1, 'x', key4, 'y'), False)
        values = yield from c.mget(key1, key2, key3, key4)
        self.assertEqual(len(values), 4)
        for value, target in zip(values, (b'1', b'2', b'3', None)):
            self.assertEqual(value, target)

    def test_set_ex(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.set(key, '1', ex=5), True)
        ttl = yield from c.ttl(key)
        self.assertTrue(0 < ttl <= 5)

    def test_set_ex_timedelta(self):
        expire_at = datetime.timedelta(seconds=5)
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.set(key, '1', ex=expire_at), True)
        ttl = yield from c.ttl(key)
        self.assertTrue(0 < ttl <= 5)

    def test_set_nx(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.set(key, '1', nx=True), True)
        yield from eq(c.set(key, '2', nx=True), False)
        yield from eq(c.get(key), b'1')

    def test_set_px(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.set(key, '1', px=10000, nx=True), True)
        yield from eq(c.get(key), b'1')
        pttl = yield from c.pttl(key)
        self.assertTrue(0 < pttl <= 10000)
        ttl = yield from c.ttl(key)
        self.assertTrue(0 < ttl <= 10)

    def test_set_px_timedelta(self):
        expire_at = datetime.timedelta(milliseconds=5000)
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.set(key, '1', px=expire_at, nx=True), True)
        yield from eq(c.get(key), b'1')
        pttl = yield from c.pttl(key)
        self.assertTrue(0 < pttl <= 5000)
        ttl = yield from c.ttl(key)
        self.assertTrue(0 < ttl <= 5)

    def test_set_xx(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.set(key, '1', xx=True), False)
        yield from eq(c.get(key), None)
        yield from eq(c.set(key, 'bar'), True)
        yield from eq(c.set(key, '2', xx=True), True)
        yield from eq(c.get(key), b'2')

    def test_setrange(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.setrange(key, 5, 'foo'), 8)
        yield from eq(c.get(key), b'\0\0\0\0\0foo')
        yield from eq(c.set(key, 'abcdefghijh'), True)
        yield from eq(c.setrange(key, 6, '12345'), 11)
        yield from eq(c.get(key), b'abcdef12345')

    def test_strlen(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.strlen(key), 0)
        yield from eq(c.set(key, 'foo'), True)
        yield from eq(c.strlen(key), 3)

    ###########################################################################
    #    HASHES
    def test_hdel(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.hdel(key, 'f1', 'f2', 'gh'), 0)
        yield from eq(c.hmset(key, {'f1': 1, 'f2': 'hello', 'f3': 'foo'}),
                      True)
        yield from eq(c.hdel(key, 'f1', 'f2', 'gh'), 2)
        yield from eq(c.hdel(key, 'fgf'), 0)
        yield from eq(c.type(key), 'hash')
        yield from eq(c.hdel(key, 'f3'), 1)
        yield from eq(c.type(key), 'none')
        yield from self._remove_and_push(key, 0)
        yield from self.async.assertRaises(ResponseError, c.hdel, key, 'foo')

    def test_hexists(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.hexists(key, 'foo'), False)
        yield from eq(c.hmset(key, {'f1': 1, 'f2': 'hello', 'f3': 'foo'}),
                      True)
        yield from eq(c.hexists(key, 'f3'), True)
        yield from eq(c.hexists(key, 'f5'), False)
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(ResponseError, c.hexists, key,
                                           'foo')

    def test_hset_hget(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.hget(key, 'foo'), None)
        yield from eq(c.hset(key, 'foo', 4), 1)
        yield from eq(c.hget(key, 'foo'), b'4')
        yield from eq(c.hset(key, 'foo', 6), 0)
        yield from eq(c.hget(key, 'foo'), b'6')
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(
            ResponseError, c.hset, key, 'foo', 7)
        yield from self.async.assertRaises(ResponseError, c.hget, key, 'foo')
        yield from self.async.assertRaises(ResponseError, c.hmset, key, 'foo')

    def test_hgetall(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        h = {b'f1': b'1', b'f2': b'hello', b'f3': b'foo'}
        yield from eq(c.hgetall(key), {})
        yield from eq(c.hmset(key, h), True)
        yield from eq(c.hgetall(key), h)
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(ResponseError, c.hgetall, key)

    def test_hincrby(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.hincrby(key, 'foo', 1), 1)
        yield from eq(c.hincrby(key, 'foo', 2), 3)
        yield from eq(c.hincrby(key, 'foo', -1), 2)
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(
            ResponseError, c.hincrby, key, 'foo', 3)

    def test_hincrbyfloat(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.hincrbyfloat(key, 'foo', 1), 1.0)
        yield from eq(c.hincrbyfloat(key, 'foo', 2.5), 3.5)
        yield from eq(c.hincrbyfloat(key, 'foo', -1.1), 2.4)
        yield from self._remove_and_push(key)

    def test_hkeys_hvals_hlen_hmget(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        h = {b'f1': b'1', b'f2': b'hello', b'f3': b'foo'}
        yield from eq(c.hkeys(key), [])
        yield from eq(c.hvals(key), [])
        yield from eq(c.hlen(key), 0)
        yield from eq(c.hmset(key, h), True)
        keys = yield from c.hkeys(key)
        vals = yield from c.hvals(key)
        self.assertEqual(sorted(keys), sorted(h))
        self.assertEqual(sorted(vals), sorted(h.values()))
        yield from eq(c.hlen(key), 3)
        yield from eq(c.hmget(key, 'f1', 'f3', 'hj'),
                      {'f1': b'1', 'f3': b'foo', 'hj': None})
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(ResponseError, c.hkeys, key)
        yield from self.async.assertRaises(ResponseError, c.hvals, key)
        yield from self.async.assertRaises(ResponseError, c.hlen, key)
        yield from self.async.assertRaises(ResponseError, c.hmget, key,
                                           'f1', 'f2')

    def test_hsetnx(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.hsetnx(key, 'a', 'foo'), 1)
        yield from eq(c.hget(key, 'a'), b'foo')
        yield from eq(c.hsetnx(key, 'a', 'bla'), 0)
        yield from eq(c.hget(key, 'a'), b'foo')
        yield from self._remove_and_push(key)
        yield from self.async.assertRaises(ResponseError, c.hsetnx, key,
                                           'a', 'jk')

    ###########################################################################
    #    LISTS
    def test_blpop(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        bk1 = key1.encode('utf-8')
        bk2 = key2.encode('utf-8')
        eq = self.async.assertEqual
        c = self.client
        yield from self.async.assertRaises(ResponseError, c.blpop, key1, 'bla')
        yield from eq(c.rpush(key1, 1, 2), 2)
        yield from eq(c.rpush(key2, 3, 4), 2)
        yield from eq(c.blpop((key2, key1), 1), (bk2, b'3'))
        yield from eq(c.blpop((key2, key1), 1), (bk2, b'4'))
        yield from eq(c.blpop((key2, key1), 1), (bk1, b'1'))
        yield from eq(c.blpop((key2, key1), 1), (bk1, b'2'))
        yield from eq(c.blpop((key2, key1), 1), None)
        yield from eq(c.rpush(key1, '1'), 1)
        yield from eq(c.blpop(key1, 1), (bk1, b'1'))

    def test_brpop(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        bk1 = key1.encode('utf-8')
        bk2 = key2.encode('utf-8')
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.rpush(key1, 1, 2), 2)
        yield from eq(c.rpush(key2, 3, 4), 2)
        yield from eq(c.brpop((key2, key1), 1), (bk2, b'4'))
        yield from eq(c.brpop((key2, key1), 1), (bk2, b'3'))
        yield from eq(c.brpop((key2, key1), 1), (bk1, b'2'))
        yield from eq(c.brpop((key2, key1), 1), (bk1, b'1'))
        yield from eq(c.brpop((key2, key1), 1), None)
        yield from eq(c.rpush(key1, '1'), 1)
        yield from eq(c.brpop(key1, 1), (bk1, b'1'))

    def test_brpoplpush(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.rpush(key1, 1, 2), 2)
        yield from eq(c.rpush(key2, 3, 4), 2)
        yield from eq(c.brpoplpush(key1, key2), b'2')
        yield from eq(c.brpoplpush(key1, key2), b'1')
        yield from eq(c.brpoplpush(key1, key2, timeout=1), None)
        yield from eq(c.lrange(key1, 0, -1), [])
        yield from eq(c.lrange(key2, 0, -1), [b'1', b'2', b'3', b'4'])

    def test_brpoplpush_empty_string(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.rpush(key1, ''), 1)
        yield from eq(c.brpoplpush(key1, key2), b'')

    def test_lindex_llen(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.lindex(key, '0'), None)
        yield from eq(c.llen(key), 0)
        yield from eq(c.rpush(key, '1', '2', '3'), 3)
        yield from eq(c.lindex(key, '0'), b'1')
        yield from eq(c.lindex(key, '1'), b'2')
        yield from eq(c.lindex(key, '2'), b'3')
        yield from eq(c.lindex(key, '3'), None)
        yield from eq(c.llen(key), 3)
        yield from self._remove_and_sadd(key)
        yield from self.async.assertRaises(ResponseError, c.lindex, key, '1')
        yield from self.async.assertRaises(ResponseError, c.llen, key)

    def test_linsert(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.linsert(key, 'after', '2', '2.5'), 0)
        yield from eq(c.rpush(key, '1', '2', '3'), 3)
        yield from eq(c.linsert(key, 'after', '2', '2.5'), 4)
        yield from eq(c.lrange(key, 0, -1), [b'1', b'2', b'2.5', b'3'])
        yield from eq(c.linsert(key, 'before', '2', '1.5'), 5)
        yield from eq(c.lrange(key, 0, -1), [b'1', b'1.5', b'2', b'2.5', b'3'])
        yield from eq(c.linsert(key, 'before', '100', '1.5'), -1)
        yield from self.async.assertRaises(ResponseError, c.linsert, key,
                                           'banana', '2', '2.5')
        yield from self._remove_and_sadd(key)
        yield from self.async.assertRaises(ResponseError, c.linsert, key,
                                           'after', '2', '2.5')

    def test_lpop(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.lpop(key), None)
        yield from eq(c.rpush(key, 1, 2), 2)
        yield from eq(c.lpop(key), b'1')
        yield from eq(c.lpop(key), b'2')
        yield from eq(c.lpop(key), None)
        yield from eq(c.type(key), 'none')
        yield from self._remove_and_sadd(key, 0)
        yield from self.async.assertRaises(ResponseError, c.lpop, key)
        yield from self.async.assertRaises(ResponseError, c.lpush, key, 4)

    def test_lset(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.rpush(key, '1', '2', '3'), 3)
        yield from eq(c.lrange(key, 0, -1), [b'1', b'2', b'3'])
        yield from eq(c.lset(key, 1, '4'), True)
        yield from eq(c.lrange(key, 0, 2), [b'1', b'4', b'3'])

    def test_ltrim(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from eq(c.rpush(key, '1', '2', '3'), 3)
        yield from eq(c.ltrim(key, 0, 1), True)
        yield from eq(c.lrange(key, 0, -1), [b'1', b'2'])

    def test_lpushx_rpushx(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.lpushx(key, 'b'), 0)
        yield from eq(c.rpushx(key, 'b'), 0)
        yield from eq(c.lrange(key, 0, -1), [])
        yield from eq(c.lpush(key, 'a'), 1)
        yield from eq(c.lpushx(key, 'b'), 2)
        yield from eq(c.rpushx(key, 'c'), 3)
        yield from eq(c.lrange(key, 0, -1), [b'b', b'a', b'c'])
        yield from self._remove_and_sadd(key)
        yield from self.async.assertRaises(ResponseError, c.lpushx, key, 'g')

    def test_lrem(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.lrem(key, 1, 'a'), 0)
        yield from eq(c.rpush(key, 'a', 'a', 'a', 'b', 'a', 'a'), 6)
        yield from eq(c.lrem(key, 1, 'a'), 1)
        yield from eq(c.lrange(key, 0, -1), [b'a', b'a', b'b', b'a', b'a'])
        yield from eq(c.lrem(key, -1, 'a'), 1)
        yield from eq(c.lrange(key, 0, -1), [b'a', b'a', b'b', b'a'])
        yield from eq(c.lrem(key, 0, 'a'), 3)
        yield from self.async.assertRaises(ResponseError, c.lrem, key,
                                           'g', 'foo')
        yield from eq(c.lrange(key, 0, -1), [b'b'])
        yield from eq(c.lrem(key, 0, 'b'), 1)
        yield from self._remove_and_sadd(key, 0)
        yield from self.async.assertRaises(ResponseError, c.lrem, key, 1)

    def test_rpop(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.rpop(key), None)
        yield from eq(c.rpush(key, 1, 2), 2)
        yield from eq(c.rpop(key), b'2')
        yield from eq(c.rpop(key), b'1')
        yield from eq(c.rpop(key), None)
        yield from eq(c.type(key), 'none')

    def test_rpoplpush(self):
        key1 = self.randomkey()
        key2 = key1 + 'x'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.rpush(key1, 'a1', 'a2', 'a3'), 3)
        yield from eq(c.rpush(key2, 'b1', 'b2', 'b3'), 3)
        yield from eq(c.rpoplpush(key1, key2), b'a3')
        yield from eq(c.lrange(key1, 0, -1), [b'a1', b'a2'])
        yield from eq(c.lrange(key2, 0, -1), [b'a3', b'b1', b'b2', b'b3'])

    ###########################################################################
    #    SORT
    def test_sort_basic(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from c.rpush(key, '3', '2', '1', '4')
        yield from eq(c.sort(key), [b'1', b'2', b'3', b'4'])

    def test_sort_limited(self):
        key = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        yield from c.rpush(key, '3', '2', '1', '4')
        yield from eq(c.sort(key, start=1, num=2), [b'2', b'3'])

    def test_sort_by(self):
        key = self.randomkey()
        key2 = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        pipe = c.pipeline()
        pipe.mset('%s:1' % key, 8,
                  '%s:2' % key, 3,
                  '%s:3' % key, 5)
        pipe.rpush(key2, '3', '2', '1')
        res = yield from pipe.commit()
        self.assertEqual(len(res), 2)
        yield from eq(c.sort(key2, by='%s:*' % key), [b'2', b'3', b'1'])

    def test_sort_get(self):
        key = self.randomkey()
        key2 = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        pipe = c.pipeline()
        pipe.mset('%s:1' % key, 'u1',
                  '%s:2' % key, 'u2',
                  '%s:3' % key, 'u3')
        pipe.rpush(key2, '3', '2', '1')
        res = yield from pipe.commit()
        self.assertEqual(len(res), 2)
        yield from eq(c.sort(key2, get='%s:*' % key), [b'u1', b'u2', b'u3'])

    def test_sort_get_multi(self):
        key = self.randomkey()
        key2 = self.randomkey()
        c = self.client
        eq = self.async.assertEqual
        pipe = c.pipeline()
        pipe.mset('%s:1' % key, 'u1',
                  '%s:2' % key, 'u2',
                  '%s:3' % key, 'u3')
        pipe.rpush(key2, '3', '2', '1')
        res = yield from pipe.commit()
        self.assertEqual(len(res), 2)
        yield from eq(c.sort(key2, get=('%s:*' % key, '#')),
                      [b'u1', b'1', b'u2', b'2', b'u3', b'3'])
        yield from eq(c.sort(key2, get=('%s:*' % key, '#'), groups=True),
                      [(b'u1', b'1'), (b'u2', b'2'), (b'u3', b'3')])

    ###########################################################################
    #    SETS
    def test_sadd_scard(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        members = (b'1', b'2', b'3', b'2')
        yield from eq(c.sadd(key, *members), 3)
        yield from eq(c.smembers(key), set(members))
        yield from eq(c.scard(key), 3)

    def test_sdiff(self):
        key = self.randomkey()
        key2 = key + '2'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3), 3)
        yield from eq(c.sdiff(key, key2), set((b'1', b'2', b'3')))
        yield from eq(c.sadd(key2, 2, 3), 2)
        yield from eq(c.sdiff(key, key2), set([b'1']))

    def test_sdiffstore(self):
        key = self.randomkey()
        key2 = key + '2'
        des = key + 'd'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3), 3)
        yield from eq(c.sdiffstore(des, key, key2), 3)
        yield from eq(c.smembers(des), set([b'1', b'2', b'3']))
        yield from eq(c.sadd(key2, 2, 3), 2)
        yield from eq(c.sdiffstore(des, key, key2), 1)
        yield from eq(c.smembers(des), set([b'1']))

    def test_sinter(self):
        key = self.randomkey()
        key2 = key + '2'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3), 3)
        yield from eq(c.sinter(key, key2), set())
        yield from eq(c.sadd(key2, 2, 3), 2)
        yield from eq(c.sinter(key, key2), set([b'2', b'3']))

    def test_sinterstore(self):
        key = self.randomkey()
        key2 = key + '2'
        des = key + 'd'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3), 3)
        yield from eq(c.sinterstore(des, key, key2), 0)
        yield from eq(c.smembers(des), set())
        yield from eq(c.sadd(key2, 2, 3), 2)
        yield from eq(c.sinterstore(des, key, key2), 2)
        yield from eq(c.smembers(des), set([b'2', b'3']))

    def test_sismember(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3), 3)
        yield from eq(c.sismember(key, 1), True)
        yield from eq(c.sismember(key, 2), True)
        yield from eq(c.sismember(key, 3), True)
        yield from eq(c.sismember(key, 4), False)

    def test_smove(self):
        key = self.randomkey()
        key2 = key + '2'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.smove(key, key2, 1), False)
        yield from eq(c.sadd(key, 1, 2), 2)
        yield from eq(c.sadd(key2, 3, 4), 2)
        yield from eq(c.smove(key, key2, 1), True)
        yield from eq(c.smembers(key), set([b'2']))
        yield from eq(c.smembers(key2), set([b'1', b'3', b'4']))

    def test_spop(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3), 3)
        value = yield from c.spop(key)
        self.assertTrue(value in set([b'1', b'2', b'3']))
        yield from eq(c.smembers(key), set([b'1', b'2', b'3']) - set([value]))

    def test_srandmember(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3), 3)
        value = yield from c.srandmember(key)
        self.assertTrue(value in set((b'1', b'2', b'3')))
        yield from eq(c.smembers(key), set((b'1', b'2', b'3')))

    def test_srandmember_multi_value(self):
        s = [b'1', b'2', b'3']
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, *s), 3)
        randoms = yield from c.srandmember(key, 2)
        self.assertEqual(len(randoms), 2)
        self.assertEqual(set(randoms).intersection(s), set(randoms))

    def test_srem(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3, 4), 4)
        yield from eq(c.srem(key, 5), 0)
        yield from eq(c.srem(key, 5), 0)
        yield from eq(c.srem(key, 2, 4), 2)
        yield from eq(c.smembers(key), set([b'1', b'3']))

    def test_sunion(self):
        key = self.randomkey()
        key2 = key + '2'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3), 3)
        yield from eq(c.sunion(key, key2), set((b'1', b'2', b'3')))
        yield from eq(c.sadd(key2, 2, 3, 4), 3)
        yield from eq(c.sunion(key, key2), set((b'1', b'2', b'3', b'4')))

    def test_sunionstore(self):
        key = self.randomkey()
        key2 = key + '2'
        des = key + 'd'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.sadd(key, 1, 2, 3), 3)
        yield from eq(c.sunionstore(des, key, key2), 3)
        yield from eq(c.smembers(des), set([b'1', b'2', b'3']))
        yield from eq(c.sadd(key2, 2, 3, 4), 3)
        yield from eq(c.sunionstore(des, key, key2), 4)
        yield from eq(c.smembers(des), set([b'1', b'2', b'3', b'4']))

    ###########################################################################
    #    SORTED SETS
    def test_zadd_zcard(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key, a1=1, a2=2, a3=3), 3)
        yield from eq(c.zrange(key, 0, -1), [b'a1', b'a2', b'a3'])
        yield from eq(c.zcard(key), 3)

    def test_zcount(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key, a1=1, a2=2, a3=3), 3)
        yield from eq(c.zcount(key, '-inf', '+inf'), 3)
        yield from eq(c.zcount(key, '(1', 2), 1)
        yield from eq(c.zcount(key, '(1', '(3'), 1)
        yield from eq(c.zcount(key, 1, 3), 3)

    def test_zincrby(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key, a1=1, a2=2, a3=3), 3)
        yield from eq(c.zincrby(key, 1, 'a2'), 3.0)
        yield from eq(c.zincrby(key, 5, 'a3'), 8.0)
        yield from eq(c.zscore(key, 'a2'), 3.0)
        yield from eq(c.zscore(key, 'a3'), 8.0)
        yield from eq(c.zscore(key, 'blaaa'), None)

    def test_zinterstore_sum(self):
        des = self.randomkey()
        key1 = des + '1'
        key2 = des + '2'
        key3 = des + '3'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key1, a1=1, a2=2, a3=1), 3)
        yield from eq(c.zadd(key2, a1=2, a2=2, a3=2), 3)
        yield from eq(c.zadd(key3, a1=6, a3=5, a4=4), 3)
        yield from eq(c.zinterstore(des, (key1, key2, key3)), 2)
        yield from eq(c.zrange(des, 0, -1, withscores=True),
                      Zset(((8.0, b'a3'), (9.0, b'a1'))))

    def test_zinterstore_max(self):
        des = self.randomkey()
        key1 = des + '1'
        key2 = des + '2'
        key3 = des + '3'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key1, a1=1, a2=2, a3=1), 3)
        yield from eq(c.zadd(key2, a1=2, a2=2, a3=2), 3)
        yield from eq(c.zadd(key3, a1=6, a3=5, a4=4), 3)
        yield from eq(c.zinterstore(des, (key1, key2, key3),
                                    aggregate='max'), 2)
        yield from eq(c.zrange(des, 0, -1, withscores=True),
                      Zset(((5.0, b'a3'), (6.0, b'a1'))))

    def test_zinterstore_min(self):
        des = self.randomkey()
        key1 = des + '1'
        key2 = des + '2'
        key3 = des + '3'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key1, a1=1, a2=2, a3=1), 3)
        yield from eq(c.zadd(key2, a1=2, a2=2, a3=2), 3)
        yield from eq(c.zadd(key3, a1=6, a3=5, a4=4), 3)
        yield from eq(c.zinterstore(des, (key1, key2, key3),
                                    aggregate='min'), 2)
        yield from eq(c.zrange(des, 0, -1, withscores=True),
                      Zset(((1.0, b'a3'), (1.0, b'a1'))))

    def test_zinterstore_with_weights(self):
        des = self.randomkey()
        key1 = des + '1'
        key2 = des + '2'
        key3 = des + '3'
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key1, a1=1, a2=2, a3=1), 3)
        yield from eq(c.zadd(key2, a1=2, a2=2, a3=2), 3)
        yield from eq(c.zadd(key3, a1=6, a3=5, a4=4), 3)
        yield from eq(c.zinterstore(des, (key1, key2, key3),
                                    weights=(1, 2, 3)), 2)
        yield from eq(c.zrange(des, 0, -1, withscores=True),
                      Zset(((20.0, b'a3'), (23.0, b'a1'))))

    def test_zrange(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key, a1=1, a2=2, a3=3), 3)
        yield from eq(c.zrange(key, 0, 1), [b'a1', b'a2'])
        yield from eq(c.zrange(key, 1, 2), [b'a2', b'a3'])
        yield from eq(c.zrange(key, 0, 1, withscores=True),
                      Zset([(1, b'a1'), (2, b'a2')]))
        yield from eq(c.zrange(key, 1, 2, withscores=True),
                      Zset([(2, b'a2'), (3, b'a3')]))

    def test_zrangebyscore(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        yield from eq(c.zrangebyscore(key, 2, 4), [b'a2', b'a3', b'a4'])
        # slicing with start/num
        yield from eq(c.zrangebyscore(key, 2, 4, offset=1, count=2),
                      [b'a3', b'a4'])
        # withscores
        yield from eq(c.zrangebyscore(key, 2, 4, withscores=True),
                      Zset([(2.0, b'a2'), (3.0, b'a3'), (4.0, b'a4')]))

    def test_zrank(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        yield from eq(c.zrank(key, 'a1'), 0)
        yield from eq(c.zrank(key, 'a2'), 1)
        yield from eq(c.zrank(key, 'a6'), None)

    def test_zrem(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        yield from eq(c.zrem(key, 'a2'), 1)
        yield from eq(c.zrange(key, 0, -1), [b'a1', b'a3', b'a4', b'a5'])
        yield from eq(c.zrem(key, 'b'), 0)
        yield from eq(c.zrange(key, 0, -1), [b'a1', b'a3', b'a4', b'a5'])
        yield from eq(c.zrem(key, 'a3', 'a5', 'h'), 2)
        yield from eq(c.zrange(key, 0, -1), [b'a1', b'a4'])
        yield from eq(c.zrem(key, 'a1', 'a4'), 2)
        yield from eq(c.type(key), 'none')

    def test_zremrangebyrank(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        yield from eq(c.zremrangebyrank(key, 1, 3), 3)
        yield from eq(c.zrange(key, 0, 5), [b'a1', b'a5'])

    def test_zremrangebyscore(self):
        key = self.randomkey()
        eq = self.async.assertEqual
        c = self.client
        yield from eq(c.zadd(key, a1=1, a2=2, a3=3, a4=4, a5=5), 5)
        yield from eq(c.zremrangebyscore(key, 2, 4), 3)
        yield from eq(c.zrange(key, 0, -1), [b'a1', b'a5'])
        yield from eq(c.zremrangebyscore(key, 2, 4), 0)
        yield from eq(c.zrange(key, 0, -1), [b'a1', b'a5'])

    ###########################################################################
    #    CONNECTION
    def test_ping(self):
        result = yield from self.client.ping()
        self.assertTrue(result)

    def test_echo(self):
        result = yield from self.client.echo('Hello')
        self.assertEqual(result, b'Hello')

    ###########################################################################
    #    SERVER
    def test_dbsize(self):
        yield from self.client.set('one_at_least', 'foo')
        result = yield from self.client.dbsize()
        self.assertTrue(result >= 1)

    def test_info(self):
        info = yield from self.client.info()
        self.assertTrue(info)
        self.assertIsInstance(info, dict)

    def test_time(self):
        t = yield from self.client.time()
        self.assertIsInstance(t, tuple)
        total = t[0] + 0.000001*t[1]
        self.assertTrue(total)

    ###########################################################################
    #    PUBSUB
    def test_handler(self):
        client = self.client
        pubsub = client.pubsub()
        self.assertEqual(client.store, pubsub.store)
        self.assertEqual(client.store._loop, pubsub._loop)
        self.assertEqual(pubsub._connection, None)

    def test_subscribe_one(self):
        key = self.randomkey()
        pubsub1 = self.client.pubsub()
        self.assertFalse(pubsub1._connection)
        # Subscribe to one channel
        yield from pubsub1.subscribe(key)
        count = yield from pubsub1.count(key)
        self.assertEqual(len(count), 1)
        self.assertEqual(count[key.encode('utf-8')], 1)
        #
        pubsub2 = self.client.pubsub()
        yield from pubsub2.subscribe(key)
        count = yield from pubsub1.count(key)
        self.assertEqual(len(count), 1)
        self.assertEqual(count[key.encode('utf-8')], 2)

    def test_subscribe_many(self):
        base = self.randomkey()
        key1 = base + '_a'
        key2 = base + '_b'
        key3 = base + '_c'
        key4 = base + 'x'
        pubsub = self.client.pubsub()
        yield from pubsub.subscribe(key1, key2, key3, key4)
        channels = yield from pubsub.channels(base + '_*')
        self.assertEqual(len(channels), 3)
        count = yield from pubsub.count(key1, key2, key3)
        self.assertEqual(len(count), 3)
        self.assertEqual(count[key1.encode('utf-8')], 1)
        self.assertEqual(count[key2.encode('utf-8')], 1)
        self.assertEqual(count[key3.encode('utf-8')], 1)

    def test_publish(self):
        pubsub = self.client.pubsub()
        listener = Listener()
        pubsub.add_client(listener)
        yield from pubsub.subscribe('chat')
        result = yield from pubsub.publish('chat', 'Hello')
        self.assertTrue(result >= 0)
        channel, message = yield from listener.get()
        self.assertEqual(channel, 'chat')
        self.assertEqual(message, b'Hello')

    def test_pattern_subscribe(self):
        # switched off for redis. Issue #95
        if self.store.name == 'pulsar':
            eq = self.async.assertEqual
            pubsub = self.client.pubsub(protocol=StringProtocol())
            listener = Listener()
            pubsub.add_client(listener)
            yield from eq(pubsub.psubscribe('f*'), None)
            yield from eq(pubsub.publish('foo', 'hello foo'), 1)
            channel, message = yield from listener.get()
            self.assertEqual(channel, 'foo')
            self.assertEqual(message, 'hello foo')
            yield from eq(pubsub.punsubscribe(), None)
            # yield from listener.get()

    ###########################################################################
    #    TRANSACTION
    def test_watch(self):
        key1 = self.randomkey()
        # key2 = key1 + '2'
        result = yield from self.client.watch(key1)
        self.assertEqual(result, 1)


class TestPulsarStore(RedisCommands, unittest.TestCase):
    app_cfg = None

    @classmethod
    def setUpClass(cls):
        server = PulsarDS(name=cls.__name__.lower(),
                          bind='127.0.0.1:0',
                          concurrency=cls.cfg.concurrency,
                          redis_py_parser=cls.redis_py_parser)
        cls.app_cfg = yield from pulsar.send('arbiter', 'run', server)
        cls.pulsards_uri = 'pulsar://%s:%s' % cls.app_cfg.addresses[0]
        cls.store = cls.create_store('%s/9' % cls.pulsards_uri)
        cls.client = cls.store.client()

    @classmethod
    def tearDownClass(cls):
        if cls.app_cfg is not None:
            return pulsar.send('arbiter', 'kill_actor', cls.app_cfg.name)

    def test_store_methods(self):
        store = self.create_store('%s/8' % self.pulsards_uri)
        self.assertEqual(store.database, 8)
        store.database = 10
        self.assertEqual(store.database, 10)
        self.assertTrue(store.dns.startswith('%s/10?' % self.pulsards_uri))
        self.assertEqual(store.encoding, 'utf-8')
        self.assertTrue(repr(store))


@unittest.skipUnless(pulsar.HAS_C_EXTENSIONS, 'Requires cython extensions')
class TestPulsarStorePyParser(TestPulsarStore):
    redis_py_parser = True



'''Tests the test suite and pulsar distribution.'''
import unittest

import pulsar
from pulsar import asyncio, send, multi_async
from pulsar.apps.test import TestSuite
from pulsar.apps.test.plugins import profile
from pulsar.utils.version import get_version


def simple_function(actor):
    return actor.name


def wait(actor, period=0.5):
    start = actor._loop.time()
    yield from asyncio.sleep(period)
    return actor._loop.time() - start


class TestTestWorker(unittest.TestCase):

    def test_TestSuiteMonitor(self):
        arbiter = pulsar.get_actor()
        self.assertTrue(len(arbiter.monitors) >= 1)
        monitor = arbiter.registered['test']
        app = monitor.app
        self.assertTrue(isinstance(app, TestSuite))

    def test_unknown_send_target(self):
        # The target does not exists
        yield from self.async.assertRaises(pulsar.CommandError, send,
                                           'vcghdvchdgcvshcd', 'ping')

    def test_multiple_execute(self):
        m = yield from multi_async((send('arbiter', 'run', wait, 1.2),
                                    send('arbiter', 'ping'),
                                    send('arbiter', 'echo', 'ciao!'),
                                    send('arbiter', 'run', wait, 2.1),
                                    send('arbiter', 'echo', 'ciao again!')))
        self.assertTrue(m[0] >= 1.1)
        self.assertEqual(m[1], 'pong')
        self.assertEqual(m[2], 'ciao!')
        self.assertTrue(m[3] >= 2.0)
        self.assertEqual(m[4], 'ciao again!')

    def test_no_plugins(self):
        suite = TestSuite()
        self.assertFalse(suite.cfg.plugins)
        self.assertFalse('profile' in suite.cfg.settings)

    def test_profile_plugins(self):
        suite = TestSuite(plugins=[profile.Profile()])
        self.assertTrue(suite.cfg.plugins)
        self.assertTrue('profile' in suite.cfg.settings)
        self.assertTrue('profile_stats_path' in suite.cfg.settings)

    def test_version(self):
        self.assertTrue(pulsar.VERSION)
        self.assertTrue(pulsar.__version__)
        self.assertEqual(pulsar.__version__, get_version(pulsar.VERSION))
        self.assertTrue(len(pulsar.VERSION) >= 2)

    def test_meta(self):
        for m in ("__author__", "__contact__", "__homepage__", "__doc__"):
            self.assertTrue(getattr(pulsar, m, None))



'''Tests the rpc middleware and utilities. It uses the calculator example.'''
import unittest

from pulsar.apps import rpc
from pulsar.apps.test import HttpTestClient


class rpcTest(unittest.TestCase):

    def proxy(self):
        from examples.calculator.manage import Site
        http = HttpTestClient(self, Site())
        return rpc.JsonProxy('http://127.0.0.1:8060/', http=http, timeout=20)

    def test_proxy(self):
        p = self.proxy()
        http = p._http
        self.assertTrue(len(http.headers))
        self.assertEqual(http.headers['user-agent'], 'Pulsar-Http-Test-Client')
        self.assertEqual(http.test, self)
        self.assertTrue(http.server_consumer)
        self.assertTrue(http.wsgi_handler)
        self.assertEqual(p._version, '2.0')



import unittest

from pulsar.utils.httpurl import Headers, SimpleCookie


class TestHeaders(unittest.TestCase):

    def testServerHeader(self):
        h = Headers()
        self.assertEqual(h.kind, 'server')
        self.assertEqual(len(h), 0)
        h['content-type'] = 'text/html'
        self.assertEqual(len(h), 1)

    def testHeaderBytes(self):
        h = Headers(kind=None)
        h['content-type'] = 'text/html'
        h['server'] = 'bla'
        self.assertTrue(repr(h).startswith('both '))
        self.assertEqual(bytes(h), b'Server: bla\r\n'
                                   b'Content-Type: text/html\r\n\r\n')

    def testClientHeader(self):
        h = Headers(kind='client')
        self.assertEqual(h.kind, 'client')
        self.assertEqual(len(h), 0)
        h['content-type'] = 'text/html'
        self.assertEqual(h.get_all('content-type'), ['text/html'])
        self.assertEqual(len(h), 1)
        h['server'] = 'bla'
        self.assertEqual(len(h), 1)
        del h['content-type']
        self.assertEqual(len(h), 0)
        self.assertEqual(h.get_all('content-type', []), [])

    def test_non_standard_request_headers(self):
        h = Headers(kind='client')
        h['accept'] = 'text/html'
        self.assertEqual(len(h), 1)
        h['server'] = 'bla'
        self.assertEqual(len(h), 1)
        h['proxy-connection'] = 'keep-alive'
        self.assertEqual(len(h), 2)
        headers = str(h)
        self.assertTrue('Proxy-Connection:' in headers)

    def test_multiple_entry(self):
        h = Headers([('Connection', 'Keep-Alive'),
                     ('Accept-Encoding', 'identity'),
                     ('Accept-Encoding', 'deflate'),
                     ('Accept-Encoding', 'compress'),
                     ('Accept-Encoding', 'gzip')],
                    kind='client')
        accept = h['accept-encoding']
        self.assertEqual(accept, 'identity, deflate, compress, gzip')

    def test_init_int(self):
        h = Headers(kind=1)
        self.assertEqual(h.kind, 'server')
        self.assertEqual(h.kind_number, 1)
        h = Headers(kind=0)
        self.assertEqual(h.kind, 'client')
        self.assertEqual(h.kind_number, 0)
        h = Headers(kind=56)
        self.assertEqual(h.kind, 'both')
        self.assertEqual(h.kind_number, 2)

    def test_remove_header(self):
        h = Headers([('Content-type', 'text/html')])
        self.assertEqual(len(h), 1)
        self.assertEqual(h.remove_header('foo'), None)
        self.assertEqual(h.remove_header('content-length'), None)
        self.assertEqual(h.remove_header('content-type'), ['text/html'])
        self.assertEqual(len(h), 0)

    def test_remove_header_value(self):
        h = Headers([('Accept-encoding', 'gzip'),
                     ('Accept-encoding', 'deflate'),
                     ('Accept', '*/*')], kind=2)
        self.assertEqual(len(h), 2)
        self.assertEqual(h['accept-encoding'], 'gzip, deflate')
        self.assertEqual(h.remove_header('accept-encoding', 'x'), None)
        self.assertEqual(h['accept-encoding'], 'gzip, deflate')
        self.assertEqual(h.remove_header('accept-encoding', 'deflate'),
                         'deflate')
        self.assertEqual(len(h), 2)
        self.assertEqual(h['accept-encoding'], 'gzip')

    def test_override(self):
        h = Headers([('Accept-encoding', 'gzip'),
                     ('Accept-encoding', 'deflate'),
                     ('Accept', '*/*')], kind=2)
        h.override([('Accept-encoding', 'gzip2'),
                    ('Accept-encoding', 'deflate2'),
                    ('Accept', 'text/html'),
                    ('Accept', '*/*; q=0.8')])
        self.assertEqual(len(h), 2)
        self.assertEqual(h['accept-encoding'], 'gzip2, deflate2')
        self.assertEqual(h['accept'], 'text/html, */*; q=0.8')

    def test_cookies(self):
        h = Headers()
        cookies = SimpleCookie({'bla': 'foo', 'pippo': 'pluto'})
        self.assertEqual(len(cookies), 2)
        for c in cookies.values():
            v = c.OutputString()
            h.add_header('Set-Cookie', v)
        h = str(h)
        self.assertTrue(
            h in ('Set-Cookie: bla=foo\r\nSet-Cookie: pippo=pluto\r\n\r\n',
                  'Set-Cookie: pippo=pluto\r\nSet-Cookie: bla=foo\r\n\r\n'))



'''Tests the test suite loader'''
import unittest


class TestFailures(unittest.TestCase):

    @unittest.expectedFailure
    def test_fail(self):
        self.assertEqual(1, 0, "broken")


class TestSetupFailure(unittest.TestCase):

    @unittest.expectedFailure
    def setUp(self):
        self.assertEqual(1, 0, "broken")

    def test_ok(self):
        # Never goes in here
        pass


class TestTearDownFailure(unittest.TestCase):
    processed = 0

    @unittest.expectedFailure
    def test_fail(self):
        self.__class__.processed += 1
        self.assertEqual(1, 0, "broken")

    def test_ok(self):
        self.__class__.processed += 1

    @classmethod
    def tearDownClass(cls):
        assert cls.processed == 2, "Should have processed 2"



import unittest

from pulsar import EventHandler, get_event_loop


class Handler(EventHandler):

    def __init__(self, **kw):
        self._loop = get_event_loop()
        super().__init__(self._loop, **kw)


class TestFailure(unittest.TestCase):

    def test_one_time(self):
        h = Handler(one_time_events=('start', 'finish'))
        h.bind_event('finish', lambda f, exc=None: 'OK')
        result = yield from h.fire_event('finish', 'foo')
        self.assertTrue(h.event('finish').done())
        self.assertEqual(result, 'foo')

    def test_one_time_error(self):
        h = Handler(one_time_events=('start', 'finish'))
        h.bind_event('finish', lambda f, exc=None: 'OK'+4)
        result = yield from h.fire_event('finish', 3)
        self.assertTrue(h.event('finish').done())
        self.assertEqual(result, 3)

    def test_bind_events(self):
        h = Handler(one_time_events=('start', 'finish'))
        h.bind_events(foo=3, bla=6)
        self.assertFalse(h.events['start'].handlers)
        self.assertFalse(h.events['finish'].handlers)
        h.bind_events(start=lambda r, exc=None: r+1,
                      finish=lambda r, exc=None: r+1)
        self.assertTrue(h.events['start'].handlers)
        self.assertTrue(h.events['finish'].handlers)
        result = yield from h.fire_event('start', 2)
        self.assertEqual(result, 2)

    def test_remove_callback(self):
        h = Handler(one_time_events=('start', 'finish'))

        def cbk(_, **kw):
            return kw

        h.bind_event('many', cbk)
        self.assertTrue(h.event('many'))
        self.assertEqual(h.remove_callback('bla', cbk), None)
        self.assertEqual(h.remove_callback('many', cbk), 1)
        self.assertEqual(h.remove_callback('many', cbk), 0)
        self.assertEqual(h.event('many').handlers, [])



'''Tests the wsgi middleware in pulsar.apps.wsgi'''
import time
import pickle
import unittest
from unittest import mock
from datetime import datetime, timedelta
from urllib.parse import urlparse

import pulsar
from pulsar.apps import wsgi
from pulsar.apps import http
from pulsar.apps.wsgi.utils import cookie_date


class WsgiRequestTests(unittest.TestCase):

    def request(self, **kwargs):
        environ = wsgi.test_wsgi_environ(**kwargs)
        return wsgi.WsgiRequest(environ)

    def test_absolute_path(self):
        uri = 'http://bbc.co.uk/news/'
        request = self.request(path=uri)
        self.assertEqual(request.get('RAW_URI'), uri)
        self.assertEqual(request.path, '/news/')
        self.assertEqual(request.absolute_uri(), uri)

    def test_is_secure(self):
        request = self.request(secure=True)
        self.assertTrue(request.is_secure)
        self.assertEqual(request.environ['HTTPS'], 'on')
        self.assertEqual(request.environ['wsgi.url_scheme'], 'https')

    def test_get_host(self):
        request = self.request(headers=[('host', 'blaa.com')])
        self.assertEqual(request.get_host(), 'blaa.com')

    def test_full_path(self):
        request = self.request(headers=[('host', 'blaa.com')])
        self.assertEqual(request.full_path(), '/')
        self.assertEqual(request.full_path('/foo'), '/foo')

    def test_full_path_query(self):
        request = self.request(path='/bla?path=foo&id=5')
        self.assertEqual(request.path, '/bla')
        self.assertEqual(request.url_data, {'path': 'foo', 'id': '5'})
        self.assertEqual(request.full_path(), '/bla')
        # self.assertTrue(request.full_path() in ('/bla?path=foo&id=5',
        #                                         '/bla?id=5&path=foo'))
        self.assertEqual(request.full_path(g=7), '/bla?g=7')

    def test_url_handling(self):
        target = '/\N{SNOWMAN}'
        request = self.request(path=target)
        path = urlparse(request.path).path
        self.assertEqual(path, target)

    def testResponse200(self):
        r = wsgi.WsgiResponse(200)
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content, ())
        self.assertFalse(r.is_streamed)
        self.assertFalse(r.started)
        self.assertEqual(list(r), [])
        self.assertTrue(r.started)
        self.assertEqual(str(r), r.status)
        self.assertTrue(repr(r))

    def testResponse500(self):
        r = wsgi.WsgiResponse(500, content=b'A critical error occurred')
        self.assertEqual(r.status_code, 500)
        self.assertEqual(r.status, '500 Internal Server Error')
        self.assertEqual(r.content, (b'A critical error occurred',))
        self.assertFalse(r.is_streamed)
        self.assertFalse(r.started)
        self.assertEqual(list(r), [b'A critical error occurred'])
        self.assertTrue(r.started)

    def testStreamed(self):
        stream = ('line {0}\n'.format(l+1) for l in range(10))
        r = wsgi.WsgiResponse(content=stream)
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.content, stream)
        self.assertTrue(r.is_streamed)
        data = []
        for l, a in enumerate(r):
            data.append(a)
            self.assertTrue(r.started)
            self.assertEqual(a, ('line {0}\n'.format(l+1)).encode('utf-8'))
        self.assertEqual(len(data), 10)

    def testForCoverage(self):
        r = wsgi.WsgiResponse(environ={'PATH_INFO': 'bla/'})
        self.assertEqual(r.path, 'bla/')
        self.assertEqual(r.connection, None)
        self.assertEqual(r.content, ())
        self.assertEqual(list(r), [])
        self.assertRaises(RuntimeError, list, r)

    def test_parse_authorization_header(self):
        parse = wsgi.parse_authorization_header
        self.assertEqual(parse(''), None)
        self.assertEqual(parse('csdcds'), None)
        self.assertEqual(parse('csdcds cbsdjchbjsc'), None)
        self.assertEqual(parse('basic cbsdjcbsjchbsd'), None)
        auths = http.HTTPBasicAuth('pippo', 'pluto').header()
        self.assertTrue(parse(auths).authenticated({}, 'pippo', 'pluto'))

    def testCookies(self):
        response = wsgi.WsgiResponse()
        expires = datetime.now() + timedelta(seconds=3600)
        response.set_cookie('bla', expires=expires)
        self.assertTrue('bla' in response.cookies)

    def testDeleteCookie(self):
        response = wsgi.WsgiResponse()
        response.delete_cookie('bla')
        self.assertTrue('bla' in response.cookies)

    def test_far_expiration(self):
        "Cookie will expire when an distant expiration time is provided"
        response = wsgi.WsgiResponse()
        response.set_cookie('datetime', expires=datetime(2028, 1, 1, 4, 5, 6))
        datetime_cookie = response.cookies['datetime']
        self.assertEqual(datetime_cookie['expires'],
                         'Sat, 01-Jan-2028 04:05:06 GMT')

    def test_max_age_expiration(self):
        "Cookie will expire if max_age is provided"
        response = wsgi.WsgiResponse()
        response.set_cookie('max_age', max_age=10)
        max_age_cookie = response.cookies['max_age']
        self.assertEqual(max_age_cookie['max-age'], 10)
        self.assertEqual(max_age_cookie['expires'],
                         cookie_date(time.time()+10))

    def test_httponly_cookie(self):
        response = wsgi.WsgiResponse()
        response.set_cookie('example', httponly=True)
        example_cookie = response.cookies['example']
        # A compat cookie may be in use -- check that it has worked
        # both as an output string, and using the cookie attributes
        self.assertTrue('; httponly' in str(example_cookie).lower())
        self.assertTrue(example_cookie['httponly'])

    def test_headers(self):
        response = wsgi.WsgiResponse(200)
        response['content-type'] = 'text/plain'
        self.assertTrue('content-type' in response)
        self.assertTrue(response.has_header('content-type'))
        self.assertEqual(response['content-type'], 'text/plain')

    def testBuildWsgiApp(self):
        appserver = wsgi.WSGIServer()
        self.assertEqual(appserver.name, 'wsgi')
        self.assertEqual(appserver.cfg.callable, None)

    def testWsgiHandler(self):
        hnd = wsgi.WsgiHandler(middleware=(wsgi.authorization_middleware,))
        self.assertEqual(len(hnd.middleware), 1)
        hnd2 = pickle.loads(pickle.dumps(hnd))
        self.assertEqual(len(hnd2.middleware), 1)

    def testHttpBinServer(self):
        from examples.httpbin.manage import server
        app = server(bind='127.0.0.1:0')
        pickle.loads(pickle.dumps(app))

    def test_clean_path_middleware(self):
        url = 'bla//foo'
        try:
            wsgi.clean_path_middleware({'PATH_INFO': url,
                                        'QUERY_STRING': 'page=1'}, None)
        except pulsar.HttpRedirect as e:
            url = e.headers[0][1]
            self.assertEqual(url, '/bla/foo?page=1')

    def test_handle_wsgi_error(self):

        def handler(request, exc):
            return 'exception: %s' % exc

        environ = wsgi.test_wsgi_environ(
            extra={'error.handler': handler})
        try:
            raise ValueError('just a test')
        except ValueError as exc:
            response = wsgi.handle_wsgi_error(environ, exc)
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.content, (b'exception: just a test',))

    def test_handle_wsgi_error_debug(self):
        cfg = self.cfg.copy()
        cfg.set('debug', True)
        environ = wsgi.test_wsgi_environ(extra={'pulsar.cfg': cfg})
        try:
            raise ValueError('just a test for debug wsgi error handler')
        except ValueError as exc:
            response = wsgi.handle_wsgi_error(environ, exc)
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.content_type, None)
        self.assertEqual(len(response.content), 1)

    def test_handle_wsgi_error_debug_html(self):
        cfg = self.cfg.copy()
        cfg.set('debug', True)
        headers = [('Accept', '*/*')]
        environ = wsgi.test_wsgi_environ(extra={'pulsar.cfg': cfg},
                                         headers=headers)
        try:
            raise ValueError('just a test for debug wsgi error handler')
        except ValueError as exc:
            response = wsgi.handle_wsgi_error(environ, exc)
        self.assertEqual(response.status_code, 500)
        html = response.content[0]
        self.assertEqual(response.content_type, 'text/html')
        self.assertTrue(html.startswith(b'<!DOCTYPE html>'))
        self.assertTrue(b'<title>500 Internal Server Error</title>' in html)

    def test_wsgi_handler_404(self):
        start = mock.MagicMock()
        handler = wsgi.WsgiHandler()
        environ = self.request().environ
        response = handler(environ, start)
        self.assertEqual(response.status_code, 404)
        self.assertEqual(start.call_count, 1)

    def test_request_redirect(self):
        request = self.request()
        response = request.redirect('/foo')
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['location'], '/foo')
        response = request.redirect('/foo2', permanent=True)
        self.assertEqual(response.status_code, 301)
        self.assertEqual(response['location'], '/foo2')



import unittest

from pulsar import get_event_loop


class TestDns(unittest.TestCase):

    def test_getnameinfo(self):
        loop = get_event_loop()
        info = yield from loop.getaddrinfo('www.bbc.co.uk', 'http')
        info = yield from loop.getnameinfo(('212.58.244.66', 80))
        info = yield from loop.getaddrinfo('github.com', 'https')
        self.assertTrue(info)
        addr = info[0][4]
        info = yield from loop.getnameinfo(addr)
        self.assertTrue(info)



import unittest
from unittest import mock

from pulsar import Future, send, multi_async, get_event_loop
from pulsar.apps import wsgi

from examples.echo.manage import server, Echo

try:
    from pulsar.apps import greenio
    run_in_greenlet = greenio.run_in_greenlet
except ImportError:
    greenio = None

    def run_in_greenlet(x):
        return x


def raise_error():
    raise RuntimeError


class EchoGreen(Echo):
    '''An echo client which uses greenlets to provide implicit
    asynchronous code'''

    def __call__(self, message):
        connection = greenio.wait(self.pool.connect())
        with connection:
            consumer = connection.current_consumer()
            consumer.start(message)
            greenio.wait(consumer.on_finished)
            return consumer if self.full_response else consumer.buffer


@unittest.skipUnless(greenio, 'Requires the greenlet package')
class TestGreenIO(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        s = server(name=cls.__name__.lower(), bind='127.0.0.1:0',
                   concurrency=cls.cfg.concurrency)
        cls.server_cfg = yield from send('arbiter', 'run', s)
        cls.client = EchoGreen(cls.server_cfg.addresses[0])

    @classmethod
    def tearDownClass(cls):
        if cls.server_cfg:
            return send('arbiter', 'kill_actor', cls.server_cfg.name)

    def request(self, **kwargs):
        environ = wsgi.test_wsgi_environ(**kwargs)
        return wsgi.WsgiRequest(environ)

    def test_pool(self):
        pool = greenio.GreenPool()
        self.assertTrue(pool._loop)
        self.assertEqual(pool._loop, get_event_loop())
        self.assertFalse(pool._greenlets)
        future = pool.submit(lambda: 'Hi!')
        self.assertIsInstance(future, Future)
        result = yield from future
        self.assertEqual(result, 'Hi!')
        self.assertEqual(len(pool._greenlets), 1)
        self.assertEqual(len(pool._available), 1)

    def test_greenlet_methods(self):
        pool = greenio.GreenPool()
        self.assertFalse(pool.in_green_worker)
        self.assertFalse(pool.getcurrent().parent)

        def _greenlet_methods():
            self.assertTrue(pool.in_green_worker)
            self.assertTrue(pool.getcurrent().parent)

        yield from pool.submit(_greenlet_methods)

    def test_error_in_pool(self):
        # Test an error
        pool = greenio.GreenPool()
        yield from self.async.assertRaises(RuntimeError, pool.submit,
                                           raise_error)
        self.assertEqual(len(pool._greenlets), 1)
        self.assertEqual(len(pool._available), 1)

    @run_in_greenlet
    def test_echo(self):
        result = self.client(b'ciao luca')
        self.assertEqual(result, b'ciao luca')

    @run_in_greenlet
    def test_large(self):
        '''Echo a 3MB message'''
        msg = b''.join((b'a' for x in range(2**13)))
        result = self.client(msg)
        self.assertEqual(result, msg)

    def test_shutdown(self):
        # Test an error
        pool = greenio.GreenPool()
        self.assertEqual(pool._max_workers, greenio._DEFAULT_WORKERS)
        yield from self.async.assertEqual(pool.submit(lambda: 'OK'), 'OK')
        self.assertEqual(len(pool._greenlets), 1)
        self.assertEqual(len(pool._available), 1)
        a = pool.submit(lambda: 'a')
        b = pool.submit(lambda: 'b')
        self.assertEqual(len(pool._greenlets), 2)
        self.assertEqual(len(pool._available), 0)
        result = yield from multi_async([a, b])
        self.assertEqual(result[0], 'a')
        self.assertEqual(result[1], 'b')
        self.assertEqual(len(pool._greenlets), 2)
        self.assertEqual(len(pool._available), 2)
        yield from pool.shutdown()
        self.assertEqual(len(pool._greenlets), 0)
        self.assertEqual(len(pool._available), 0)

    def test_lock_error(self):
        lock = greenio.GreenLock()
        self.assertFalse(lock.locked())
        self.assertRaises(RuntimeError, lock.acquire)
        self.assertFalse(lock.locked())
        self.assertRaises(RuntimeError, lock.release)

    @run_in_greenlet
    def test_lock(self):
        green = greenio.getcurrent()
        lock = greenio.GreenLock()
        self.assertTrue(lock.acquire())
        self.assertEqual(lock.locked(), green)

        def _test_lock(l):
            return l.acquire()
        #
        # create a new greenlet
        child = greenio.greenlet(_test_lock)
        future = child.switch(lock)

        self.assertIsInstance(future, Future)
        self.assertEqual(lock.locked(), green)

        # release the lock
        lock.release()
        self.assertTrue(future.done())
        self.assertEqual(lock.locked(), green)
        #
        # self.assertEqual(lock.locked(), child)

    def test_greenwsgi(self):
        from pulsar.apps.greenio.wsgi import GreenWSGI
        wsgi = mock.MagicMock()
        pool = greenio.GreenPool()
        green = GreenWSGI(wsgi, pool)
        self.assertEqual(green.wsgi, wsgi)
        self.assertEqual(green.pool, pool)

    def test_uncatched_stopiteration(self):
        pool = greenio.GreenPool()
        with self.assertRaises(RuntimeError) as cm:
            yield from pool.submit(lambda: next(iter([])))
        self.assertIsInstance(cm.exception.__cause__, StopIteration)



# -*- coding: utf-8 -*-

import unittest

from pulsar.utils.slugify import slugify


class TestSlugify(unittest.TestCase):

    def test_manager(self):

        txt = "This is a test ---"
        r = slugify(txt)
        self.assertEqual(r, "this-is-a-test")

        txt = "This -- is a ## test ---"
        r = slugify(txt)
        self.assertEqual(r, "this-is-a-test")

        txt = 'C\'est déjà l\'été.'
        r = slugify(txt)
        self.assertEqual(r, "cest-deja-lete")

        txt = 'Nín hǎo. Wǒ shì zhōng guó rén'
        r = slugify(txt)
        self.assertEqual(r, "nin-hao-wo-shi-zhong-guo-ren")

        txt = 'Компьютер'
        r = slugify(txt)
        self.assertEqual(r, "kompiuter")

        txt = 'jaja---lol-méméméoo--a'
        r = slugify(txt)
        self.assertEqual(r, "jaja-lol-mememeoo-a")

        txt = 'jaja---lol-méméméoo--a'
        r = slugify(txt, max_length=9)
        self.assertEqual(r, "jaja-lol")

        txt = 'jaja---lol-méméméoo--a'
        r = slugify(txt, max_length=15)
        self.assertEqual(r, "jaja-lol-mememe")

        txt = 'jaja---lol-méméméoo--a'
        r = slugify(txt, max_length=50)
        self.assertEqual(r, "jaja-lol-mememeoo-a")

        txt = 'jaja---lol-méméméoo--a'
        r = slugify(txt, max_length=15, word_boundary=True)
        self.assertEqual(r, "jaja-lol-a")

        txt = 'jaja---lol-méméméoo--a'
        r = slugify(txt, max_length=19, word_boundary=True)
        self.assertEqual(r, "jaja-lol-mememeoo")

        txt = 'jaja---lol-méméméoo--a'
        r = slugify(txt, max_length=20, word_boundary=True)
        self.assertEqual(r, "jaja-lol-mememeoo-a")

        txt = 'jaja---lol-méméméoo--a'
        r = slugify(txt, max_length=20, word_boundary=True, separator=".")
        self.assertEqual(r, "jaja.lol.mememeoo.a")

        txt = 'jaja---lol-méméméoo--a'
        r = slugify(txt, max_length=20, word_boundary=True, separator="ZZZZZZ")
        self.assertEqual(r, "jajaZZZZZZlolZZZZZZmememeooZZZZZZa")



from random import choice
import string
import unittest

from pulsar import HAS_C_EXTENSIONS
from pulsar.apps.ds import redis_parser

characters = string.ascii_letters + string.digits


class RedisPyParser(unittest.TestCase):
    __benchmark__ = True
    __number__ = 100
    _sizes = {'tiny': 2,
              'small': 10,
              'normal': 100,
              'big': 1000,
              'huge': 10000}
    redis_py_parser = True

    @classmethod
    def setUpClass(cls):
        size = cls.cfg.size
        nsize = cls._sizes[size]
        cls.data = [''.join((choice(characters) for l in range(20)))
                    for s in range(nsize)]
        cls.data_bytes = [(''.join((choice(characters) for l in range(20)))
                           ).encode('utf-8') for s in range(nsize)]
        cls.parser = redis_parser(cls.redis_py_parser)()
        cls.chunk = cls.parser.multi_bulk(cls.data)

    def test_pack_command(self):
        self.parser.pack_command(self.data)

    def test_encode_multi_bulk(self):
        self.parser.multi_bulk(self.data)

    def test_encode_multi_bulk_bytes(self):
        self.parser.multi_bulk(self.data_bytes)

    def test_decode_multi_bulk(self):
        self.parser.feed(self.chunk)
        self.parser.get()


@unittest.skipUnless(HAS_C_EXTENSIONS, 'Requires C extensions')
class RedisCParser(RedisPyParser):
    redis_py_parser = False



import unittest

from pulsar import send
from pulsar.apps import greenio

from examples.echo.manage import server, Echo
from tests.apps.greenio import EchoGreen


class TestGreenIo(unittest.TestCase):
    __benchmark__ = True
    __number__ = 1000

    @classmethod
    def setUpClass(cls):
        s = server(name=cls.__name__.lower(), bind='127.0.0.1:0')
        cls.server_cfg = yield from send('arbiter', 'run', s)
        cls.client = Echo(cls.server_cfg.addresses[0])
        cls.green = EchoGreen(cls.server_cfg.addresses[0])
        cls.msg = b''.join((b'a' for x in range(2**13)))
        cls.pool = greenio.GreenPool()

    def test_yield_io(self):
        result = yield from self.client(self.msg)
        self.assertEqual(result, self.msg)

    @greenio.run_in_greenlet
    def test_green_io(self):
        result = self.green(self.msg)
        self.assertEqual(result, self.msg)

    def test_green_pool(self):
        result = yield from self.pool.submit(self.green, self.msg)
        self.assertEqual(result, self.msg)



import asyncio
from functools import partial
from collections import namedtuple
from copy import copy
from urllib.parse import urlparse, urljoin

from pulsar import OneTime, is_async
from pulsar.apps.ws import WebSocketProtocol, WS
from pulsar.utils.httpurl import REDIRECT_CODES, requote_uri, SimpleCookie
from pulsar.utils.websocket import SUPPORTED_VERSIONS, websocket_key
from pulsar import PulsarException


def noerror(callback):
    '''Decorator to run a callback of a :class:`.EventHandler`
    only when no errors occur
    '''
    def _(*response, **kw):
        if response[-1] and not kw.get('exc'):
            return callback(*response)

    return _


def response_content(resp, exc=None, **kw):
    b = resp.parser.recv_body()
    if b or resp._content is None:
        resp._content = resp._content + b if resp._content else b
    return resp._content


def _consumer(response, consumer):
    if response is not None:
        consumer = response
    return consumer


def start_request(request, conn):
    response = conn.current_consumer()
    # bind request-specific events
    response.bind_events(**request.inp_params)
    if request.auth:
        response.bind_event('pre_request', request.auth)

    if request.stream:
        response.bind_event('data_processed', response.raw)
        response.start(request)
        yield from response.events['on_headers']
    else:
        response.bind_event('data_processed', response_content)
        response.start(request)
        yield from response.on_finished

    if hasattr(response.request_again, '__call__'):
        response = response.request_again(response)
        if is_async(response):
            response = yield from response

    return response


class request_again(namedtuple('request_again', 'method url params')):

    @property
    def status_code(self):
        return -1

    @property
    def headers(self):
        return ()


class TooManyRedirects(PulsarException):

    def __init__(self, response):
        self.response = response


class WebSocketClient(WebSocketProtocol):
    status_code = 101

    @property
    def _request(self):
        return self.handshake._request

    @property
    def headers(self):
        return self.handshake.headers

    def __getattr__(self, name):
        if not name.startswith('__'):
            return getattr(self.handshake, name)
        else:
            raise AttributeError("'%s' object has no attribute '%s'" %
                                 (self.__class__.__name__, name))


class Redirect:

    @noerror
    def __call__(self, response, exc=None):
        if (response.status_code in REDIRECT_CODES and
                'location' in response.headers and
                response.request.allow_redirects):
            response.request_again = self._do_redirect

    def _do_redirect(self, response):
        request = response.request
        # done with current response
        url = response.headers.get('location')
        # Handle redirection without scheme (see: RFC 1808 Section 4)
        if url.startswith('//'):
            parsed_rurl = urlparse(request.full_url)
            url = '%s:%s' % (parsed_rurl.scheme, url)
        # Facilitate non-RFC2616-compliant 'location' headers
        # (e.g. '/path/to/resource' instead of
        # 'http://domain.tld/path/to/resource')
        if not urlparse(url).netloc:
            url = urljoin(request.full_url,
                          # Compliant with RFC3986, we percent
                          # encode the url.
                          requote_uri(url))
        history = request.history
        if history and len(history) >= request.max_redirects:
            raise TooManyRedirects(response)

        params = request.inp_params.copy()
        params['history'] = copy(history) if history else []
        params['history'].append(response)
        if response.status_code == 303:
            method = 'GET'
            params.pop('data', None)
            params.pop('files', None)
        else:
            method = request.method
        response.request_again = request_again(method, url, params)
        return response


@noerror
def handle_cookies(response, exc=None):
    '''Handle response cookies.
    '''
    headers = response.headers
    request = response.request
    client = request.client
    response._cookies = c = SimpleCookie()
    if 'set-cookie' in headers or 'set-cookie2' in headers:
        for cookie in (headers.get('set-cookie2'),
                       headers.get('set-cookie')):
            if cookie:
                c.load(cookie)
        if client.store_cookies:
            client.cookies.extract_cookies(response, request)


@noerror
def handle_100(response, exc=None):
    '''Handle Except: 100-continue.

    This is a ``on_header`` hook which checks if the request headers
    have the ``Expect: 100-continue`` value. If so add a ``on_headers``
    callback to handle the response from the server.
    '''
    request = response.request
    if (request.headers.has('expect', '100-continue') and
            response.status_code == 100):
        response.bind_event('on_headers', _write_body)


@noerror
def _write_body(response, exc=None):
    if response.status_code == 100:
        response.request.new_parser()
        response.write_body()


class WebSocket:

    @property
    def websocket_key(self):
        if not hasattr(self, '_websocket_key'):
            self._websocket_key = websocket_key()
        return self._websocket_key

    @noerror
    def __call__(self, response, exc=None):
        request = response.request
        if request and request.scheme in ('ws', 'wss'):
            headers = request.headers
            headers['connection'] = 'Upgrade'
            headers['upgrade'] = 'websocket'
            if 'Sec-WebSocket-Version' not in headers:
                headers['Sec-WebSocket-Version'] = str(max(SUPPORTED_VERSIONS))
            if 'Sec-WebSocket-Key' not in headers:
                headers['Sec-WebSocket-Key'] = self.websocket_key
            response.bind_event('on_headers', self.on_headers)

    @noerror
    def on_headers(self, response, exc=None):
        '''Websocket upgrade as ``on_headers`` event.'''

        if response.status_code == 101:
            connection = response.connection
            request = response.request
            handler = request.websocket_handler
            if not handler:
                handler = WS()
            parser = request.client.frame_parser(kind=1)
            consumer = partial(WebSocketClient, response, handler, parser)
            connection.upgrade(consumer)
            body = response.recv_body()
            response.finished()
            websocket = connection.current_consumer()
            websocket.data_received(body)
            response.request_again = lambda r: websocket


class Tunneling:
    '''A pre request callback for handling proxy tunneling.

    If Tunnelling is required, it writes the CONNECT headers and abort
    the writing of the actual request until headers from the proxy server
    are received.
    '''
    def __init__(self, loop):
        assert loop
        self._loop = loop

    @noerror
    def __call__(self, response, exc=None):
        # the pre_request handler
        request = response.request
        if request and request.tunnel:
            request.tunnel.apply(response, self)

    @noerror
    def on_headers(self, response, exc=None):
        '''Called back once the headers have arrived.'''
        if response.status_code == 200:
            response.request_again = self._tunnel_request
            response.finished()

    def _tunnel_request(self, response):
        request = response.request.request
        connection = response.connection
        loop = connection._loop
        sock = connection.sock
        connection.transport.pause_reading()
        yield None
        # set a new connection_made event
        connection.events['connection_made'] = OneTime(loop=loop)
        connection._processed -= 1
        connection.producer._requests_processed -= 1
        #
        # For some reason when using the code below, it fails in python 3.5
        # _, connection = yield from loop._create_connection_transport(
        #         sock, lambda: connection,
        #         request._ssl,
        #         server_hostname=request._netloc)
        #
        # Therefore use legacy SSL transport
        waiter = asyncio.Future(loop=loop)
        loop._make_legacy_ssl_transport(sock, connection, request._ssl,
                                        waiter,
                                        server_hostname=request._netloc)
        yield from waiter
        #
        yield from connection.event('connection_made')
        response = yield from start_request(request, connection)
        return response



import os
import unittest

from pulsar import send
from pulsar.apps.test import ActorTestMixin
from pulsar.utils.tools import Pidfile


class TestPidfile(ActorTestMixin, unittest.TestCase):
    concurrency = 'process'

    def test_create_pid(self):
        proxy = yield from self.spawn_actor(name='pippo')
        info = yield from send(proxy, 'info')
        result = info['actor']
        self.assertTrue(result['is_process'])
        pid = result['process_id']
        #
        p = Pidfile()
        self.assertEqual(p.fname, None)
        self.assertEqual(p.pid, None)
        p.create(pid)
        self.assertTrue(p.fname)
        self.assertEqual(p.pid, pid)
        p1 = Pidfile(p.fname)
        self.assertRaises(RuntimeError, p1.create, p.pid+1)
        #
        p1 = Pidfile('bla/ksdcskcbnskcdbskcbksdjcb')
        self.assertRaises(RuntimeError, p1.create, p.pid+1)
        p1.unlink()
        p.unlink()
        self.assertFalse(os.path.exists(p.fname))



import requests

from pulsar.apps.test import dont_run_with_thread

__test__ = False


class TestRequest:
    session = requests.session()

    @dont_run_with_thread
    def test_requests_get_200(self):
        http = self.session
        response = http.get(self.httpbin(), verify=False,
                            proxies=self.proxies())
        self.assertEqual(str(response), '<Response [200]>')
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.content)
        self.assertEqual(response.url, self.httpbin())
        response = http.get(self.httpbin('get'), verify=False,
                            proxies=self.proxies())
        self.assertEqual(response.status_code, 200)



from pulsar import arbiter, spawn, send, ensure_future, Config, command


PREFIX = 'remote_'


@command()
def remote_call(request, cls, method, args, kw):
    '''Command for executing remote calls on a remote object
    '''
    actor = request.actor
    name = 'remote_%s' % cls.__name__
    if not hasattr(actor, name):
        object = cls(actor)
        setattr(actor, name, object)
    else:
        object = getattr(actor, name)
    method_name = '%s%s' % (PREFIX, method)
    return getattr(object, method_name)(request, *args, **kw)


class RemoteType(type):
    '''Metaclass for remote objects.

    Collect a set of method prefixed with ``remote_``
    '''
    def __new__(cls, name, bases, attrs):
        remotes = set()
        for base in bases[::-1]:
            if hasattr(base, 'remote_methods'):
                remotes.update(base.remote_methods)

        for key, method in list(attrs.items()):
            if hasattr(method, '__call__') and key.startswith(PREFIX):
                method_name = key[len(PREFIX):]
                remotes.add(method_name)
        for base in bases[::-1]:
            if hasattr(base, 'remote_methods'):
                remotes.update(base.remote_methods)

        attrs['remote_methods'] = frozenset(remotes)
        return super().__new__(cls, name, bases, attrs)


class RemoteCall:
    __slots__ = ('remote', 'name')

    def __init__(self, remote, name):
        self.remote = remote
        self.name = name

    def __call__(self, *args, **kwargs):
        return send(self.remote.proxy, 'remote_call', self.remote.__class__,
                    self.name, args, kwargs)


class Remote(metaclass=RemoteType):
    '''Base class for Remote Objects

    Subclass your remote object from this class and add `remote_*`
    methods which are executed in the remote actor.
    '''
    proxy = None

    def __init__(self, proxy):
        self.proxy = proxy

    @classmethod
    def spawn(cls, **kwargs):
        proxy = yield from spawn(**kwargs)
        return cls(proxy)

    def __getattr__(self, name):
        if name in self.remote_methods:
            return RemoteCall(self, name)
        else:
            return getattr(self.proxy, name)


# EXAMPLE APPLICATION

class Calculator(Remote):
    '''A simple example
    '''
    value = 0

    def remote_set_value(self, request, value):
        request.actor.logger.info('Setting value')
        self.value = value

    def remote_get_value(self, request):
        request.actor.logger.info('Getting value')
        return self.value


def start(arbiter, **kw):
    ensure_future(app(arbiter))


def app(arbiter):
    # Spawn a new actor
    calc = yield from Calculator.spawn(name='calc1')
    print(calc.name)
    # set value in the remote calculator
    yield from calc.set_value(46)
    # get value from the remote calculator
    value = yield from calc.get_value()
    print(value)

    # Stop the application
    arbiter.stop()


if __name__ == '__main__':
    cfg = Config()
    cfg.parse_command_line()
    arbiter(cfg=cfg, start=start).start()



from pulsar.apps import http
import requests     # noqa

if __name__ == '__main__':
    session = http.HttpClient()
    # session = requests.session()
    #
    url = 'https://api.github.com/'
    verify = True
    #
    url = 'https://127.0.0.1:8070'
    verify = False
    #
    proxies = {'https': 'http://127.0.0.1:8060'}
    response = session.get(url, proxies=proxies, verify=verify)
    print(response)
    # proxies = {'https': 'http://127.0.0.1:8080'}
    # response = session.get(url, proxies=proxies, verify=verify)
    # print(response)
    proxies = {'https': 'http://127.0.0.1:8060'}
    response = session.get(url, proxies=proxies, verify=verify)
    print(response)



import socket
import unittest

from pulsar import get_actor

from .base import TestHttpClientBase


class ExternalBase(TestHttpClientBase):
    with_httpbin = False

    def after_response(self, response):
        pass

    def ___test_http_get_timeit(self):
        client = self.client()
        N = 20
        responses = yield from client.timeit(N, 'get', 'http://www.bbc.co.uk/')
        self.assertEqual(len(responses), N)
        for n in range(N):
            all = []

            def save_data(r, data=None):
                all.append(data)
            try:
                response = yield from client.get('http://www.theguardian.com/',
                                                 data_received=save_data)
            except Exception:
                for n, d in enumerate(all):
                    with open('data%s.dat' % n, 'wb') as f:
                        f.write(d)
                raise
            self.assertEqual(response.status_code, 200)

    def __test_http_get(self):
        client = self.client()
        response = yield from client.get('http://www.bbc.co.uk/')
        self.assertEqual(response.status_code, 200)
        self.after_response(response)

    def test_get_https(self):
        client = self.client()
        response = yield from client.get('https://github.com/trending')
        self.assertEqual(response.status_code, 200)

    def __test_bad_host(self):
        client = self.client()
        try:
            response = yield from client.get('http://xxxyyyxxxxyyy/blafoo')
        except socket.error:
            pass
        else:
            self.assertTrue(response.request.proxy)
            self.assertTrue(response.status_code >= 400)


class ProxyExternal(ExternalBase):

    def after_response(self, response):
        self.assertTrue(response.request.proxy)

    def test_get_https(self):
        client = self.client()
        response = yield from client.get('https://github.com/trending')
        self.assertEqual(response.status_code, 200)


@unittest.skipUnless(get_actor().cfg.http_proxy == '',
                     'Requires no external proxy')
class Test_HttpClient_NoProxy_External(ExternalBase, unittest.TestCase):
    '''Test external URI when no global proxy server is present.
    '''


@unittest.skipUnless(get_actor().cfg.http_proxy == '',
                     'Requires no external proxy')
class Test_HttpClient_Proxy_External(ProxyExternal, unittest.TestCase):
    with_proxy = True


@unittest.skipUnless(get_actor().cfg.http_proxy, 'Requires external proxy')
class Test_HttpClient_ExternalProxy_External(ProxyExternal,
                                             unittest.TestCase):
    pass



import asyncio

from pulsar import isfuture, ensure_future


class StreamConsumedError(Exception):
    '''Same name as request error'''
    pass


class HttpStream:
    '''A streaming body for an HTTP response
    '''
    def __init__(self, response):
        self._streamed = False
        self._response = response
        self._queue = asyncio.Queue()

    def __repr__(self):
        return repr(self._response)
    __str__ = __repr__

    @property
    def done(self):
        return self._response.on_finished.fired()

    def read(self, n=None):
        '''Read all content'''
        if self._streamed:
            return b''
        buffer = []
        for body in self:
            if isfuture(body):
                body = yield from body
            buffer.append(body)
        return b''.join(buffer)

    def close(self):
        pass

    def __iter__(self):
        '''Iterator over bytes or Futures resulting in bytes
        '''
        if self._streamed:
            raise StreamConsumedError
        self._streamed = True
        self(self._response)
        while True:
            if self.done:
                try:
                    yield self._queue.get_nowait()
                except asyncio.QueueEmpty:
                    break
            else:
                yield ensure_future(self._queue.get())

    def __call__(self, response, exc=None, **kw):
        if self._streamed and response.parser.is_headers_complete():
            assert response is self._response
            self._queue.put_nowait(response.recv_body())



import unittest
from random import randint

from pulsar.utils.structures import Zset
from pulsar.apps.test import populate


class TestZset(unittest.TestCase):
    zset = Zset

    def random(self):
        string = populate('string', size=100)
        values = populate('float', size=100, min=-10, max=10)
        s = self.zset()
        s.update(zip(values, string))
        return s

    def test_add(self):
        s = self.zset()
        s.add(3, 'ciao')
        s.add(4, 'bla')
        self.assertEqual(len(s), 2)
        s.add(-1, 'bla')
        self.assertEqual(len(s), 2)
        data = list(s)
        self.assertEqual(data, ['bla', 'ciao'])

    def test_rank(self):
        s = self.zset()
        s.add(3, 'ciao')
        s.add(4, 'bla')
        s.add(2, 'foo')
        s.add(20, 'pippo')
        s.add(-1, 'bla')
        self.assertEqual(len(s), 4)
        self.assertEqual(s.rank('bla'), 0)
        self.assertEqual(s.rank('foo'), 1)
        self.assertEqual(s.rank('ciao'), 2)
        self.assertEqual(s.rank('pippo'), 3)
        self.assertEqual(s.rank('xxxx'), None)

    def test_update(self):
        s = self.random()
        self.assertTrue(s)
        prev = None
        for score, _ in s.items():
            if prev is not None:
                self.assertTrue(score >= prev)
            prev = score
        return s

    def test_remove(self):
        s = self.test_update()
        values = list(s)
        while values:
            index = randint(0, len(values)-1)
            val = values.pop(index)
            self.assertTrue(val in s)
            self.assertNotEqual(s.remove(val), None)
            self.assertFalse(val in s)
        self.assertFalse(s)

    def test_remove_same_score(self):
        s = self.zset([(3, 'bla'), (3, 'foo'), (3, 'pippo')])
        self.assertEqual(s.remove('foo'), 3)
        self.assertEqual(len(s), 2)
        self.assertFalse('foo' in s)

    def test_range(self):
        s = self.random()
        values = list(s.range(3, 10))
        self.assertTrue(values)
        self.assertEqual(len(values), 7)
        all = list(s)[3:10]
        self.assertEqual(all, values)

    def test_range_scores(self):
        s = self.random()
        values = list(s.range(3, 10, True))
        self.assertTrue(values)
        self.assertEqual(len(values), 7)
        all = list(s)[3:10]
        all2 = [v for _, v in values]
        self.assertEqual(all, all2)

    def test_remove_range_by_score(self):
        s = self.zset([(1.2, 'bla'), (2.3, 'foo'), (3.6, 'pippo')])
        self.assertEqual(s.remove_range_by_score(1.6, 4), 2)
        self.assertEqual(s, self.zset([(1.2, 'bla')]))

    def test_remove_range_by_rank(self):
        s = self.zset([(1.2, 'bla'), (2.3, 'foo'), (3.6, 'pippo'),
                       (4, 'b'), (5, 'c')])
        self.assertEqual(s.remove_range(1, 4), 3)
        self.assertEqual(s, self.zset([(1.2, 'bla'), (5, 'c')]))



import unittest

from pulsar.utils import security
from pulsar.utils.string import random_string


class TestSecurity(unittest.TestCase):

    def testSalt(self):
        s1 = security.gen_salt(10)
        self.assertEqual(len(s1), 10)
        self.assertNotEqual(security.gen_salt(10), s1)
        s1 = security.gen_salt(30)
        self.assertEqual(len(s1), 30)
        self.assertRaises(ValueError, security.gen_salt, 0)

    def testPassword(self):
        password = 'my-test$$-password'
        hash = security.generate_password_hash(password)
        self.assertTrue('$' in hash)
        self.assertFalse(security.check_password_hash('bla', 'bla'))
        self.assertFalse(security.check_password_hash(hash, 'bla'))
        self.assertFalse(security.check_password_hash(hash, 'bla$foo'))
        self.assertTrue(security.check_password_hash(hash, password))

    def test_random_string(self):
        s1 = random_string(min_len=20)
        self.assertEqual(len(s1), 20)
        self.assertIsInstance(s1, str)
        self.assertNotEqual(s1, random_string(min_len=20))
        self.assertNotEqual(s1, random_string(min_len=20))
        self.assertNotEqual(s1, random_string(min_len=20))
